import datetime
import random
import textwrap
import warnings
from typing import Iterable

import langchain_core
import langsmith.evaluation
import langsmith.schemas
import numpy as np
import prefect
import pydantic
import tqdm.auto

from ai_app.ai_utils import to_json_llm_input, try_converting_to_openai_flex_tier
from ai_app.apps.jira_service_desk.backend import Bot, get_system_prompt
from ai_app.apps.jira_service_desk.data import (
    JiraRequest,
    JiraRequestPrimaryKey,
    ManualJiraRequestGuide,
    fetch_use_case_guides_for_requests,
    get_saved_request_type_primary_keys,
)
from ai_app.apps.jira_service_desk.preparation import generate_requests
from ai_app.config import get_ai_postgres_engine, get_chat_model, get_judge_model_names
from ai_app.core import Response
from ai_app.external.atlassian import build_issue_url
from ai_app.utils import wrap_with_xml_tag


class JiraServiceDeskBotEvaluationExample(pydantic.BaseModel):
    generated_user_query: str
    issue_url: str
    model_used_for_generation: str
    request_type_primary_key: JiraRequestPrimaryKey
    issue_key: str


def generate_evaluation_examples_from_latest_resolved_requests(
    request_primary_key: JiraRequestPrimaryKey,
    limit_requests: int = 1,
) -> Iterable[JiraServiceDeskBotEvaluationExample]:
    requests = list(
        generate_requests(**request_primary_key.model_dump(), limit_requests=limit_requests)
    )
    if len(requests) != limit_requests:
        warnings.warn(
            f"Failed to fetch {limit_requests} requests for primary key {request_primary_key}, "
            f"instead fetched {len(requests)} requests."
        )
        return

    for request in requests:
        issue_key = request["Issue key"]
        judge_model_name = random.choice(get_judge_model_names())
        model = get_chat_model(judge_model_name)
        model = try_converting_to_openai_flex_tier(model)
        system_prompt = textwrap.dedent("""
            We are evaluating a Jira service desk support bot designed to help users find the correct 
            request type and assist with creating and filling out Jira requests. To test the bot, we 
            need realistic synthetic user messages that could have been sent to the bot before a request
            was created.

            **Your task**:

            - Review the provided resolved Jira request in JSON format.
            - Generate a message that a user might have sent to the bot before submitting this request.
            - The message should be natural and reflect how real users communicate: it may contain 
                errors, typos, be brief or incomplete, and written in a hurry.
            - Assume the user knows they are chatting with a bot, so the message is likely direct and 
                to the point, not polite or formal.
            - The message should be in the same language as the request description or comments.
            
            **Produce a single, realistic user message that could have initiated the provided Jira 
            request.**
        """)  # noqa: E501
        response = model.invoke(
            [
                langchain_core.messages.SystemMessage(system_prompt),
                langchain_core.messages.HumanMessage(to_json_llm_input(request)),
            ]
        )
        example = JiraServiceDeskBotEvaluationExample(
            request_type_primary_key=request_primary_key,
            issue_key=issue_key,
            issue_url=build_issue_url(issue_key=issue_key),
            generated_user_query=response.content,
            model_used_for_generation=judge_model_name,
        )
        yield example


def get_or_create_jira_service_desk_dataset(
    dataset_name: str | None = None,
    limit_requests: int = 1,
) -> langsmith.schemas.Dataset:
    dataset_name = dataset_name or f"Jira service desk {datetime.datetime.now():%Y-%m-%d %H:%M:%S}"
    client = langsmith.Client()
    if client.has_dataset(dataset_name=dataset_name):
        dataset = client.read_dataset(dataset_name=dataset_name)
        return dataset

    keys = get_saved_request_type_primary_keys(
        get_ai_postgres_engine(),
        JiraRequest.get_sa_primary_key() == ManualJiraRequestGuide.get_sa_primary_key(),
        ManualJiraRequestGuide.do_recommend,
    )
    examples = []
    for key in tqdm.auto.tqdm(keys):
        for example in generate_evaluation_examples_from_latest_resolved_requests(
            key, limit_requests=limit_requests
        ):
            example = langsmith.schemas.ExampleCreate(
                inputs=example.model_dump(),
                metadata=example.model_dump(include=["model_used_for_generation", "issue_url"]),
            )
            examples.append(example)

    dataset = client.create_dataset(dataset_name=dataset_name)
    client.create_examples(dataset_name=dataset_name, examples=examples)
    return dataset


class JiraServiceDeskBotEvaluation(pydantic.BaseModel):
    evaluation: str = pydantic.Field(
        description=textwrap.dedent("""
            Evaluation of the Jira service desk bot response according to the system prompt.
            All text information should be in English, concise, clear, dry, and to the point.
        """),
    )
    score: float = pydantic.Field(
        description="The score of the Jira service desk bot response, from 0 to 10.",
        ge=0,
        le=10,
    )
    updated_use_case_summary: str | None = pydantic.Field(
        None,
        description=textwrap.dedent("""
            If the bot correctly recommended the request type, then None.
            Otherwise, an updated use case summary, that should include the user's problem, while
            keeping most of content from the original summary.
        """),
    )
    system_prompt_improvement: str | None = pydantic.Field(
        None,
        description=textwrap.dedent("""
            If the bot's system prompt can be meaningfully improved, then provide a concise 
            improvement suggestion.
        """),
    )
    insight: str = pydantic.Field(
        description=textwrap.dedent("""
            A short, actionable insight or conclusion in English following the bot evaluation, for 
            example:
            - If the bot failed to recommend correct request type, why, and how to fix it.
            - How the bot behavior or its context can be improved.
        """),
    )


class Evaluator:
    def __init__(self, judge_model_names: list[str] | None = None):
        self.judge_model_names = judge_model_names or get_judge_model_names()

    def evaluate_example(
        self, inputs: dict, outputs: dict, reference_outputs: dict
    ) -> langsmith.evaluation.EvaluationResults | langsmith.evaluation.EvaluationResult:
        example = JiraServiceDeskBotEvaluationExample(**inputs)
        requests = generate_requests(
            **example.request_type_primary_key.model_dump(),
            jql_filter=f"issue = {example.issue_key}",
        )
        requests = list(requests)
        if len(requests) != 1:
            raise RuntimeError(f"Failed to fetch the Jira request {example.issue_key}.")

        request = requests[0]
        guides = fetch_use_case_guides_for_requests([example.request_type_primary_key])
        if len(guides) > 1:
            raise RuntimeError(
                f"Expected to fetch zero or one guide for request type "
                f"{example.request_type_primary_key}, but fetched {len(guides)} guides."
            )

        guide = None if not guides else guides[0]
        context = {
            "Jira request": request,
            "Use case guide for the Jira request type": guide,
            # "Generated user query": example.generated_user_query,
            "Jira service desk bot conversation": outputs,
        }

        jira_service_desk_system_prompt = wrap_with_xml_tag(
            "jira_service_desk_system_prompt", get_system_prompt(), with_new_lines=True
        )
        system_prompt = textwrap.dedent("""
            # Role
            You are an expert quality assurance specialist evaluating a Jira Service Desk bot.
            Your assessment will help improve the bot's performance and user experience.
            
            # Context
            For evaluation purposes, a real resolved Jira request was taken, and a synthetic user
            message was generated based on this request, with which the user may have come to the
            Jira service desk bot. Then the Jira service desk bot generated a response to this message.
            You will be provided with the following context in JSON format:
            1. The real resolved Jira request.
            2. The complete human and AI conversation with one turn each in LangChain messages format,
            complete with tool calls and responses.
            3. The use case guide for the Jira request type that the bot relies on when deciding 
            which request type to suggest. If the use case guide is of low quality or doesn't cover
            the provided request, it may explain why the bot failed to recommend the correct
            request type.

            # Task
            Your task is to evaluate the quality of the Jira service desk bot response based on the
            following criteria:
            1. High importance:
                - **Adherence to instructions:** The response adheres to the bot's system prompt:
                {jira_service_desk_system_prompt}
            2. Medium importance:
                - **Correctness:** If the user message was not detailed enough for a clear request type
                recommendation, then ensure that the bot asked adequate questions. Otherwise, check
                whether the actual request type is appropriate for the user's query - maybe the resolved
                request type was incorrectly created in the first place. If not, then check
                that the suggested request type is the same as the one in the real resolved request. 
                If the bot suggested a different request type, inspect the context provided to the bot -
                the RAG tool response, and try to understand why the bot failed to recommend the correct 
                request type. Attribute the failure to bot performance, the quality of generated human 
                message, or the mismatch of the RAG tool output.
            3. Low importance:
                - **Relevance:** The response is directly relevant to the user's query.
                - **Clarity:** The response is clear and easy to understand.
                - **Usefulness:** The response provides useful information to the user.
                - **Professionalism:** The response is professional and courteous.
        """).format(jira_service_desk_system_prompt=jira_service_desk_system_prompt)  # noqa: E501
        judge_model_input = [
            langchain_core.messages.SystemMessage(system_prompt),
            langchain_core.messages.HumanMessage(to_json_llm_input(context)),
        ]
        results = []
        score_values = []
        for judge_model_name in self.judge_model_names:
            judge_model = get_chat_model(judge_model_name)
            judge_model = try_converting_to_openai_flex_tier(judge_model)
            judge_model = judge_model.with_structured_output(JiraServiceDeskBotEvaluation)
            evaluation: JiraServiceDeskBotEvaluation = judge_model.invoke(judge_model_input)
            score_values.append(evaluation.score)
            score = langsmith.evaluation.EvaluationResult(
                key=f"Score by {judge_model_name}", score=evaluation.score
            )
            review = langsmith.evaluation.EvaluationResult(
                key=f"Review by {judge_model_name}",
                value="\n\n".join(f"# {k}\n{v}" for k, v in evaluation.model_dump().items()),
            )
            updated_use_case_summary = langsmith.evaluation.EvaluationResult(
                key=f"Updated use case summary by {judge_model_name}",
                value=evaluation.updated_use_case_summary,
            )
            results.extend([score, review, updated_use_case_summary])

        average_result = langsmith.evaluation.EvaluationResult(
            key="Average score", score=np.mean(score_values)
        )
        results.append(average_result)
        results = langsmith.evaluation.EvaluationResults(results=results)
        return results


async def ask_jira_bot(
    example: JiraServiceDeskBotEvaluationExample,
    model_name: str,
    n_closest_requests: int | None = None,
) -> dict:
    user_message = langchain_core.messages.HumanMessage(content=example.generated_user_query)
    bot = Bot(n_closest_requests=n_closest_requests)
    response: Response = await bot.respond(
        model=get_chat_model(model_name),
        messages=[user_message],
    )
    response = {
        # "ai_message": response.content_for_langsmith,
        "langchain_messages": "\n".join(m.pretty_repr() for m in response.messages),
    }
    return response


@prefect.task
async def evaluate_dataset(
    dataset_name: str,
    bot_model_name: str,
    n_closest_requests: int | None = None,
    judge_model_names: list[str] | None = None,
):
    async def process_inputs(inputs: dict) -> dict:
        example = JiraServiceDeskBotEvaluationExample(**inputs)
        response = await ask_jira_bot(
            example=example,
            model_name=bot_model_name,
            n_closest_requests=n_closest_requests,
        )
        return response

    evaluator = Evaluator(judge_model_names=judge_model_names)
    client = langsmith.Client()
    await client.aevaluate(
        process_inputs,
        data=dataset_name,
        evaluators=[evaluator.evaluate_example],
        metadata={
            "Bot model": bot_model_name,
            "Number of closest documents from RAG": n_closest_requests,
        },
    )



import contextlib
import logging
from typing import Iterable

import logfire
import numpy as np
import pgvector.sqlalchemy
import sqlalchemy as sa
import sqlmodel


@contextlib.contextmanager
def open_session(
    engine: sa.Engine,
    commit_on_exit: bool = False,
    use_autocommit_isolation_level: bool = False,
    logfire_span_name: str = "SQLModel session",
    **session_kwargs,
):
    if use_autocommit_isolation_level:
        engine = engine.execution_options(isolation_level="AUTOCOMMIT")

    with sqlmodel.Session(engine, **session_kwargs) as session, logfire.span(logfire_span_name):
        yield session
        if commit_on_exit:
            session.commit()


def execute_text_statements(engine: sa.Engine, commands: Iterable[str], **kwargs):
    for command in commands:
        with open_session(engine, commit_on_exit=True, **kwargs) as session:
            session.exec(sa.text(command))


def setup_pgvector(engine: sa.Engine):
    """Requires superuser permissions."""
    execute_text_statements(engine, ["CREATE EXTENSION IF NOT EXISTS vector"])


def create_vector_index(
    engine: sa.Engine,
    table_name: str,
    column_name: str = "embedding",
    index_type: str = "hnsw ({column_name} vector_cosine_ops)",
):
    index_type = index_type.format(column_name=column_name)
    statement = f"""
        CREATE INDEX CONCURRENTLY IF NOT EXISTS {table_name}_embedding_index
        ON {table_name}
        USING {index_type}
    """
    execute_text_statements(engine, [statement], use_autocommit_isolation_level=True)


def get_cosine_distance_embedding_sqlmodel_mixin(
    dimension: int, nullable: bool = False
) -> type[sqlmodel.SQLModel]:
    class CosineDistanceEmbeddingMixin(sqlmodel.SQLModel):
        embedding: list[float] | None if nullable else list[float] = sqlmodel.Field(
            sa_type=pgvector.sqlalchemy.Vector(dimension), nullable=nullable
        )

        @classmethod
        def create_vector_index(cls, engine: sa.Engine):
            create_vector_index(engine, cls.__tablename__)

        @classmethod
        def distance(cls, other: np.ndarray) -> sa.sql.elements.BinaryExpression:
            return cls.embedding.cosine_distance(other)

        @classmethod
        def nearest_neighbors(
            cls, select: sa.sql.Select, other: np.ndarray, limit: int
        ) -> sa.sql.Select:
            return select.order_by(cls.distance(other)).limit(limit)

    return CosineDistanceEmbeddingMixin


def user_exists(engine: sa.Engine, username: str) -> bool:
    """Check if a PostgreSQL user exists."""
    with engine.connect() as conn:
        statement = f"SELECT 1 FROM pg_roles WHERE rolname = '{username}'"
        result = conn.execute(sa.text(statement)).scalar()

    return bool(result)


def setup_user(engine: sa.Engine, user: str, password: str):
    if user_exists(engine, user):
        logging.info(f"User {user} already exists, skipping creation")
        return

    statements = f"""
        CREATE USER {user} WITH PASSWORD '{password}';

        GRANT USAGE ON SCHEMA public TO {user};
        GRANT CREATE ON SCHEMA public TO {user};

        GRANT SELECT, INSERT, UPDATE, DELETE ON ALL TABLES IN SCHEMA public TO {user};

        GRANT USAGE, SELECT, UPDATE ON ALL SEQUENCES IN SCHEMA public TO {user};

        ALTER DEFAULT PRIVILEGES IN SCHEMA public
        GRANT SELECT, INSERT, UPDATE, DELETE ON TABLES TO {user};

        ALTER DEFAULT PRIVILEGES IN SCHEMA public
        GRANT USAGE, SELECT, UPDATE ON SEQUENCES TO {user};
    """
    execute_text_statements(engine, [statements])


def get_column_path(column: sa.Column, with_schema: bool = False) -> str:
    path = [column.table.name, column.name]
    if with_schema:
        path = [column.table.schema] + path

    path = ".".join(path)
    return path


def drop_all_tables(engine: sa.Engine):
    metadata = sa.MetaData()
    metadata.reflect(engine)
    metadata.drop_all(engine)


def build_foreign_key_constraint(
    sql_model: type[sqlmodel.SQLModel], ondelete: str | None = "CASCADE"
) -> sa.ForeignKeyConstraint:
    primary_keys = [c for c in sql_model.__table__.c if c.primary_key]
    columns = [c.name for c in primary_keys]
    refcolumns = [f"{c.table.name}.{c.name}" for c in primary_keys]
    constraint = sa.ForeignKeyConstraint(columns, refcolumns, ondelete=ondelete)
    return constraint


def create_index(
    engine: sa.Engine,
    table_name: str,
    columns: Iterable[str],
    schema: str | None = None,
    unique: bool = False,
):
    metadata = sa.MetaData(schema=schema)
    metadata.reflect(engine, only=[table_name])
    table = f"{schema}.{table_name}" if schema else table_name
    table = metadata.tables[table]
    with engine.connect() as connection:
        index = sa.Index(
            None,  # Index name.
            *(table.c[c] for c in columns),
            unique=unique,
        )
        index.create(connection, checkfirst=True)
        connection.commit()



import os
import urllib
from typing import Iterable

import langchain_core.messages
import openai
import polars as pl
import pydantic
import requests

from ai_app.utils import unnest_with_prefix


def get_openai_api_url():
    return "https://api.openai.com"


def prepare_request(
    messages: Iterable[langchain_core.messages.BaseMessage],
    model: str = "gpt-4o-mini",
    structured_output: pydantic.BaseModel | None = None,
    strict_output: bool = False,
    method: str = "POST",
    path: str = "v1/chat/completions",
    api_key: str | None = None,
) -> requests.Request:
    """
    The structured_output is expected to have a config dict compatible with OpenAI structured output schema format.
    Returns a well-formed HTTP request that, for example, can be used with the `requests` library:
        `response = requests.request(**vars(request))`
    """
    url = urllib.parse.urljoin(get_openai_api_url(), path)
    messages = langchain_core.messages.convert_to_openai_messages(messages)
    data = {
        "messages": messages,
        "model": model,
    }
    if structured_output:
        data["response_format"] = {
            "type": "json_schema",
            "json_schema": {
                "schema": structured_output.model_json_schema(mode="serialization"),
                "name": structured_output.__name__,
                "strict": strict_output,
            },
        }

    headers = {
        "Content-Type": "application/json",
    }
    if api_key:
        headers["Authorization"] = f"Bearer {api_key}"

    request = requests.Request(method=method, url=url, headers=headers, json=data)
    return request


def write_openai_batch_api_jsonl(requests_: list[requests.Request], path: str) -> None:
    if os.path.exists(path):
        os.remove(path)

    # Requests are written one by one to prevent polars from coalescting struct columns by adding
    # None values to missing fields.
    with open(path, "a") as f:
        for index, request in enumerate(requests_):
            df = pl.DataFrame([vars(request)])
            df = df.select(
                pl.lit(str(index)).alias("custom_id"),
                pl.col("url").str.strip_prefix(get_openai_api_url()),
                "method",
                pl.col("json").alias("body"),
            )
            df.write_ndjson(f)


def create_openai_batch(
    path: str,
    client: openai.OpenAI | None = None,
    endpoint: str = "/v1/chat/completions",
    metadata: dict | None = None,
    rename_file: bool = False,
):
    client = client or openai.OpenAI()
    with open(path, "rb") as f:
        input_file = client.files.create(file=f, purpose="batch")

    if rename_file:
        directory, file = os.path.split(path)
        new_path = os.path.join(directory, input_file.id)
        os.rename(path, new_path)

    batch = client.batches.create(
        input_file_id=input_file.id,
        endpoint=endpoint,
        completion_window="24h",
        metadata=metadata,
    )
    return batch


def openai_collection_to_dataframe(
    collection: Iterable[pydantic.BaseModel], time_zone: str = "Asia/Baku"
) -> pl.DataFrame | None:
    df = pl.DataFrame([i.model_dump() for i in collection], infer_schema_length=None)
    if df.is_empty():
        return

    epoch_columns = [c for c in df.columns if c.endswith("_at") and df.schema[c].is_integer()]
    df = df.with_columns(pl.from_epoch(c).dt.convert_time_zone(time_zone) for c in epoch_columns)
    return df


def fetch_openai_files_metadata(
    client: openai.OpenAI | None = None, limit: int = 10_000, **kwargs
) -> pl.DataFrame | None:
    client = client or openai.OpenAI()
    files = client.files.list(limit=limit, **kwargs)
    df = openai_collection_to_dataframe(files)
    return df


def fetch_openai_batches_metadata(
    client: openai.OpenAI | None = None, limit: int = 100, **kwargs
) -> pl.DataFrame:
    client = client or openai.OpenAI()
    batches = client.batches.list(limit=limit, **kwargs)
    df = openai_collection_to_dataframe(batches)
    df = unnest_with_prefix(df, "request_counts")
    files = fetch_openai_files_metadata()
    if files is not None:
        files = files.select(
            pl.col("filename").alias("input_filename"),
            pl.col("id").alias("input_file_id"),
        )
        df = df.join(files, on="input_file_id", how="left")

    return df


def try_fetch_batch_output(
    batch_id: str, client: openai.OpenAI | None = None
) -> pl.DataFrame | None:
    client = client or openai.OpenAI()
    batch = client.batches.retrieve(batch_id)
    if not batch.output_file_id:
        return

    file = client.files.content(batch.output_file_id)
    df = pl.read_ndjson(file.content)
    df = df.unnest("response")
    df = unnest_with_prefix(df, "body")
    if (df["body_choices"].list.len() == 1).all():
        df = df.with_columns(pl.col("body_choices").list[0].struct["message"].struct["content"])

    return df



import inspect
import textwrap
from typing import AsyncGenerator

import bir_mcp
import langchain_core.language_models
import langchain_core.messages
import langgraph.prebuilt

from ai_app.ai_utils import (
    AstreamResponse,
    PreModelHook,
    anonymize_tool_output,
    astream_response_messages,
    trim_history_messages,
)
from ai_app.config import (
    get_anon_codec,
    get_chat_model,
    get_gitlab_toolkit,
    get_jira_toolkit,
    get_sonarqube_toolkit,
)
from ai_app.core import State
from ai_app.pii import EntityTypeCodec, get_sanitized_input_handling_guide
from ai_app.tools import (
    InfoSecUserReminder,
    extract_tool_artifacts_from_agent_response,
    report_disclosure_of_sensitive_information,
    report_feedback,
    report_masked_entities_difficulties,
)


def build_agent(
    model: str | langchain_core.language_models.BaseChatModel, codec: EntityTypeCodec | None = None
):
    model = get_chat_model(model) if isinstance(model, str) else model
    gitlab_toolkit = get_gitlab_toolkit()
    jira_toolkit = get_jira_toolkit()
    jira_tools = jira_toolkit.get_tools()
    jira_tools = [
        anonymize_tool_output(tool, anon_codec=get_anon_codec(), codec=codec) for tool in jira_tools
    ]
    sonarqube_toolkit = get_sonarqube_toolkit()
    tools = (
        gitlab_toolkit.get_tools()
        + sonarqube_toolkit.get_tools()
        + jira_tools
        + [
            # report_disclosure_of_sensitive_information,
            report_masked_entities_difficulties,
            report_feedback,
        ]
    )
    prompt = f"""
        Your job is to evaluate a GitLab merge request. You have tools to fetch details of a merge request by its url 
        and to fetch description of a Jira issue by its key, as well as other tools. Inspect the tool descriptions and call
        them whenever appropriate, preferrably in parallel. You will be provided with a message asking you to review a merge request.
        Additionally, several Jira issues may be provided related to the merge request. You should perform the following steps:
        1. Fetch the details for the GitLab merge request.
        2. If Jira issues were provided by a user or if any Jira issues were mentioned in the merge request title or description, 
        fetch their descriptions and determine whether the issues correlate with the changes in the merge request.
        3. Review the changes in the merge request, evaluate the quality of the code, check for any potential bugs and security issues.
        4. Give a final verdict, whether the merge request is ready to be merged, if not, provide reasons and suggest next actions.
        {get_sanitized_input_handling_guide(codec=codec)}
    """  # noqa: E501
    prompt = textwrap.dedent(prompt)
    prompt = bir_mcp.git_lab.prompts.get_merge_request_review_prompt()
    prompt += f"\nNote about the tool outputs:\n{get_sanitized_input_handling_guide(codec=codec)}"
    prompt += inspect.cleandoc(f"""
        Note about user feedback:
        The user will be asked to provide feedback about the conducted merge request review.
        If you detect any explicit, meaningful feedback, report it via the {report_feedback.name} 
        tool.
    """)

    agent = langgraph.prebuilt.create_react_agent(
        model,
        tools=tools,
        prompt=prompt,
        pre_model_hook=PreModelHook(model=model),
        # By default, providing response format adds an extra call to the model, which is costly.
        # Maybe can be fixed if the structured output is converted to a tool with
        # return_direct=True.
        # response_format=Response,
    )
    return agent


def get_feedback_request_prompt() -> str:
    prompt = textwrap.dedent(
        """

        *If you have any feedback or suggestions regarding this merge request review, please let me know. 
        Your feedback will be used to improve the next version of the review agent.*
        """  # noqa: E501
    )
    return prompt


class Bot:
    # async def respond(
    #     self,
    #     model: langchain_core.language_models.BaseChatModel,
    #     model_input: list[langchain_core.messages.BaseMessage],
    #     codec: EntityTypeCodec | None = None,
    # ) -> str:
    #     agent = build_agent(model, codec=codec)
    #     response: dict = await agent.ainvoke({"messages": model_input})
    #     response_message = response["messages"][-1].content
    #     artifacts = extract_tool_artifacts_from_agent_response(response["messages"])
    #     user_reminders: list[InfoSecUserReminder] | None = artifacts.get(
    #         report_disclosure_of_sensitive_information.name
    #     )
    #     if user_reminders:
    #         user_reminder = user_reminders[0]
    #         response_message = f"**{user_reminder.message}**\n\n{response_message}"

    #     if any(not isinstance(m, langchain_core.messages.AIMessage) for m in model_input):
    #         # If this is the first AI message, ask for feedback.
    #         response_message += get_feedback_request_prompt()

    #     return response_message

    async def astream(
        self,
        model: langchain_core.language_models.BaseChatModel,
        messages: list[langchain_core.messages.BaseMessage],
        state: State,
    ) -> AsyncGenerator[AstreamResponse, None]:
        messages = trim_history_messages(messages)
        agent = build_agent(model, codec=state.codec)
        response = AstreamResponse(formatted_response="", state=[])
        async for response in astream_response_messages(agent, messages, stream_ai_tokens=True):
            yield response

        if all(not isinstance(m, langchain_core.messages.AIMessage) for m in messages):
            yield AstreamResponse(
                response.formatted_response + get_feedback_request_prompt(), response.state
            )



import inspect
import textwrap

import cachetools.func
import langchain.tools
import langchain_core.messages
import langgraph.prebuilt
import pydantic

from ai_app.ai_utils import PreModelHook, astream_response_messages
from ai_app.apps.jira_service_desk.agent_graph import create_initial_agent
from ai_app.apps.jira_service_desk.data import (
    JiraRequestPrimaryKey,
    get_saved_request_type_primary_keys,
)
from ai_app.apps.jira_service_desk.tools import (
    JiraServiceDeskToolkit,
    build_url_for_jira_request_type,
    fetch_detailed_jira_request_type_guide,
)
from ai_app.config import get_ai_postgres_engine, get_config
from ai_app.core import Response, State
from ai_app.frontend import get_generating_text_span
from ai_app.tools import (
    InfoSecUserReminder,
    build_structured_response_tool,
    extract_tool_artifacts_from_agent_response,
    get_structured_output_from_agent_response,
    report_disclosure_of_sensitive_information,
    report_feedback,
    think,
)


def get_system_prompt() -> str:
    system_prompt = textwrap.dedent(f"""
        # Role
        You are a Jira service desk expert in a bank. Your goal is to help employees find the most 
        appropriate Jira request type for their problem.

        # Language instruction
        - Your primary instruction, above all else, is to detect the language of the user's very 
        last message (Azerbaijani, English, or Russian) and ALWAYS respond in that exact language.
        - DO NOT deviate from this rule, even if other context suggests a different language might be helpful.

        # Steps
        - Use the RAG retriever "{JiraServiceDeskToolkit.fetch_jira_request_type_guides_closest_to_user_query.__name__}" 
        tool to find relevant Jira request types based on the user's messages.
        - Use the "{think.__name__}" tool to plan the next step and determine whether exactly one request type clearly and unambiguously matches
        the user's problem. 
        - If there is no clear standout request type (multiple plausible matches or the message is too vague),
        then ask concise clarifying questions to gather the missing details and narrow down the request type. Otherwise:
            - Call the "{fetch_detailed_jira_request_type_guide.__name__}" tool for the best-matching request type in the same conversation turn.
            - Respond on the first line with a one-sentence recommendation that names the request type and why it fits.
            - On the second line, provide the non-prefilled request type URL from the tool output.
            - Then briefly provide and format three sections:
                - Self-help: how the user may try to resolve the issue independently.
                - Resolution pipeline: how this request is typically handled to resolution.
                - Required information: what is needed to create the request.
        - After finding a relevant request type, in subsequent conversation turns collect from the user the information needed to create 
        a prefilled URL for the suggested request type via the "{build_url_for_jira_request_type.__name__}" 
        tool and provide it to the user. Focus on required fields that can be passed as URL parameters.

        # Guidelines
        - Always call one or several tools, never respond without calling a tool.
        - Proactively ask for user's feedback and use the "{report_feedback.name}" tool to report it, 
        if user provided any explicit feedback.
        - Use the "{build_structured_response_tool(None).name}" tool to generate the message for the user.

        # Examples
        ## Example 1 (clear, unambiguous match):
        User: '''
            I need to report a fraud related to my account.
        '''
        Bot: '''
            The 'Fraud' request type can be used to report suspected fraudulent activity affecting your account or transactions.
            https://jira-support.kapitalbank.az/servicedesk/customer/portal/1/create/1

            ### Self-help
            Immediately lock your card in the mobile app and review recent transactions.
            
            ### Resolution pipeline
            Triage → Transaction review by Fraud Ops → Temporary block if needed → Investigation → Customer notification and resolution.
            
            ### Required information for creating the request
            - Affected account/card
            - Transaction IDs and timestamps
            - Merchant name/location
        '''

        ## Example 2 (ambiguous, needs clarification):
        User: '''
            My card has a problem.
        '''
        Bot: '''
            I found several possible request types 'Card Replacement', 'Card PIN Reset', 'Card Transaction Dispute', could you clarify:
            1. Is the card lost/stolen, damaged, or working but a transaction looks wrong?
            2. If a transaction issue: when did it occur and what merchant?
            3. Do you still have physical possession of the card?
        '''
    """)  # noqa: E501
    return system_prompt


def build_prompts() -> list[langchain_core.messages.BaseMessage]:
    system_prompt = get_system_prompt()
    prompts = [
        langchain_core.messages.SystemMessage(system_prompt),
    ]
    return prompts


class BotResponse(pydantic.BaseModel):
    last_user_message_language: str = pydantic.Field(
        description=inspect.cleandoc("""
            The language of the last user message, which the bot should also use for its own 
            response.
        """),
    )
    content: str = pydantic.Field(description="The message content that the user will see.")


@cachetools.func.ttl_cache(ttl=60)
def get_cached_saved_request_type_primary_keys() -> set[JiraRequestPrimaryKey]:
    return get_saved_request_type_primary_keys(get_ai_postgres_engine())


class Bot:
    def __init__(self, n_closest_requests: int | None = None):
        self.toolkit = JiraServiceDeskToolkit(
            n_closest_requests=n_closest_requests
            or get_config().apps.jira_service_desk.n_closest_requests,
            service_desks=get_config().apps.jira_service_desk.service_desks,
        )

    async def astream(self, model, messages, state: State | None = None):
        if not any(isinstance(m, langchain_core.messages.AIMessage) for m in messages):
            # Initial agent flow can be defined more strictly, since there's less entropy in
            # the conversation state. Also it has more importance and impact than subsequent
            # agent responses, which can be designed as a simple ReAct agent loop.
            agent = create_initial_agent(model, self.toolkit)
        else:
            agent = langgraph.prebuilt.create_react_agent(
                model,
                prompt=get_system_prompt(),
                pre_model_hook=PreModelHook(model=model),
                tools=self.toolkit.get_tools()
                + [
                    # report_disclosure_of_sensitive_information,
                    report_feedback,
                    build_structured_response_tool(BotResponse),
                    langchain.tools.tool(think),  # Add thinking tool for non-reasoning models.
                ],
                response_format=BotResponse,
            )

        response = None
        async for response in astream_response_messages(
            agent, messages, stream_ai_tokens=True, with_spinner=False
        ):
            yield Response(
                content=response.formatted_response + get_generating_text_span(),
                messages=response.state["messages"],
            )

        if not response:
            return

        bot_response = get_structured_output_from_agent_response(response.state)
        yield Response(
            content=response.formatted_response + bot_response.content,
            messages=response.state["messages"],
            content_for_langsmith=bot_response.content,
        )

    async def respond(self, model, messages, state: State | None = None) -> Response:
        response = Response(content="")
        async for response in self.astream(model, messages, state):
            pass

        return response



import collections
import inspect
import logging
import re
from typing import Iterable, override

import bidict
import httpx
import presidio_analyzer
import presidio_anonymizer
import pydantic

from ai_app.utils import CounterBidict


def get_analyzer_supported_entities(analyzer: presidio_analyzer.AnalyzerEngine) -> list[str]:
    """https://microsoft.github.io/presidio/supported_entities"""
    supported_entities = set()
    for recognizer in analyzer.get_recognizers():
        supported_entities |= set(recognizer.supported_entities)

    return sorted(supported_entities)


class EntityTypeCodec:
    """A statefull class, accumulating seen entities with their assigned ids."""

    def __init__(self, replacing_format: str = "<{entity_type}_{index}>"):
        self.replacing_format = replacing_format
        self.entity_type_counter = collections.defaultdict(CounterBidict)
        self.encoded_entities = bidict.bidict()

    def format_indexed_entity_type(self, entity_type: str, index: int = 0) -> str:
        text = self.replacing_format.format(entity_type=entity_type, index=index)
        return text

    def encode(self, entity_type: str, entity: str) -> str:
        """Operates on an extracted entity string."""
        index = self.entity_type_counter[entity_type][entity]
        encoded_entity = self.format_indexed_entity_type(entity_type=entity_type, index=index)
        self.encoded_entities[entity] = encoded_entity
        return encoded_entity

    def decode(self, encoded_text: str) -> str:
        """Operates on a whole encoded text with multiple encoded entities."""
        if not self.encoded_entities:
            return encoded_text

        decoded_text = re.sub(
            pattern="|".join(self.encoded_entities.values()),
            repl=lambda match: self.encoded_entities.inverse[match.group()],
            string=encoded_text,
        )
        return decoded_text


class CodecOperatorConfig(pydantic.BaseModel, arbitrary_types_allowed=True):
    entity_type: str
    codec: EntityTypeCodec

    def encode(self, entity: str) -> str:
        entity = self.codec.encode(self.entity_type, entity)
        return entity


class EntityCounterOperator(presidio_anonymizer.operators.Operator):
    @override
    @staticmethod
    def operator_name() -> str:
        return "entity_counter"

    @override
    def validate(self, params: dict) -> CodecOperatorConfig | None:
        config = CodecOperatorConfig(**params)
        return config

    @staticmethod
    def build_config(codec):
        config = presidio_anonymizer.OperatorConfig(
            EntityCounterAnonymizer.operator_name(),
            params=dict(codec=codec),
        )
        return config


class EntityCounterAnonymizer(EntityCounterOperator):
    @override
    @staticmethod
    def operator_type() -> presidio_anonymizer.operators.OperatorType:
        return presidio_anonymizer.operators.OperatorType.Anonymize

    @override
    def operate(self, text: str, params: dict) -> str:
        config = self.validate(params)
        anonimized_entity = config.encode(text)
        return anonimized_entity


def build_custom_analyzer(
    deny_list: Iterable[str] | None = None,
) -> presidio_analyzer.AnalyzerEngine:
    # Score is the confidence value from 0 to 1 that the pattern matched something
    # worth anonymizing.
    patterns = [
        presidio_analyzer.Pattern(
            name="AZ_PASSPORT", regex=r"\b(AR|AZ|AZE|AA|MYI|MYİ|MOM|QV|DYI|DYİ)\d{6,}\b", score=0.9
        ),
        presidio_analyzer.Pattern(name="KOD", regex=r"\bKOD ?\d{6,}\b", score=0.9),
        presidio_analyzer.Pattern(name="VOEN", regex=r"\b\d{10}\b", score=0.5),
        # Middle ground between masking senstive numeric ids and too many false positives.
        # Should cover ARN, RRN, account numbers, etc. Doesn't account for spaces in between digits.
        presidio_analyzer.Pattern(name="NUMERIC_ID", regex=r"\b\d{12,}\b", score=0.5),
        presidio_analyzer.Pattern(
            name="ALPHANUMERIC_ID",
            regex=(
                r"\b"
                r"(?=[A-Z0-9]{6,}\b)"  # Look ahead for at least 6 alphanumeric chars.
                r"(?=[0-9]*[A-Z])"  # Look ahead for at least 1 letter.
                r"(?=(?:[A-Z]*[0-9]){2})"  # Look ahead for at least 2 numbers using a non-capturing group.  # noqa: E501
                r"[A-Z0-9]*"  # Consume the chars.
                r"\b"
            ),
            score=0.5,
        ),
    ]
    recognizers = [
        presidio_analyzer.predefined_recognizers.CreditCardRecognizer(),
        presidio_analyzer.predefined_recognizers.CryptoRecognizer(),
        presidio_analyzer.predefined_recognizers.IbanRecognizer(),
        presidio_analyzer.predefined_recognizers.IpRecognizer(),
        presidio_analyzer.predefined_recognizers.MedicalLicenseRecognizer(),
        presidio_analyzer.predefined_recognizers.PhoneRecognizer(),
        *[presidio_analyzer.PatternRecognizer(p.name, patterns=[p]) for p in patterns],
    ]
    if deny_list:
        recognizers.append(presidio_analyzer.PatternRecognizer("SECRET", deny_list=deny_list))

    registry = presidio_analyzer.recognizer_registry.RecognizerRegistry(recognizers=recognizers)
    configuration = {
        "nlp_engine_name": "spacy",
        "models": [
            {"lang_code": "en", "model_name": "en_core_web_md"},
        ],
        "ner_model_configuration": {
            "model_to_presidio_entity_mapping": (
                presidio_analyzer.nlp_engine.ner_model_configuration.MODEL_TO_PRESIDIO_ENTITY_MAPPING
                | {"FAC": "LOCATION"}  # Turn FACility label into LOCATION.
            ),
            # LABELS_TO_IGNORE is removed from older presidio versions.
            # "labels_to_ignore": (
            #     presidio_analyzer.nlp_engine.ner_model_configuration.LABELS_TO_IGNORE
            # ),
            "low_score_entity_names": [],
        },
    }
    provider = presidio_analyzer.nlp_engine.NlpEngineProvider(nlp_configuration=configuration)
    nlp_engine = provider.create_engine()
    analyzer = presidio_analyzer.AnalyzerEngine(registry=registry, nlp_engine=nlp_engine)
    return analyzer


class AnonCodec:
    """
    Note that the presence of EntityTypeCodec attribute makes this class stateful when not passing
    custom codec into methods.
    """

    def __init__(
        self,
        analyzer: presidio_analyzer.AnalyzerEngine | None = None,
        codec: EntityTypeCodec | None = None,
        language: str = "en",
        deny_list: Iterable[str] | None = None,
        debug: bool = False,
    ):
        self.analyzer = analyzer or build_custom_analyzer(deny_list=deny_list)
        self.codec = codec or EntityTypeCodec()
        self.language = language
        self.debug = debug
        self.anon_engine = presidio_anonymizer.AnonymizerEngine()
        self.anon_engine.add_anonymizer(EntityCounterAnonymizer)

    def anonymize(
        self, text: str, language: str | None = None, codec: EntityTypeCodec | None = None
    ) -> str:
        codec = codec or EntityTypeCodec()
        analyzer_results = self.analyzer.analyze(text=text, language=language or self.language)
        result = self.anon_engine.anonymize(
            text,
            analyzer_results,
            {"DEFAULT": EntityCounterOperator.build_config(codec)},
        )
        anon_text = result.text
        if self.debug:
            deanon_text = self.deanonymize(anon_text, codec=codec)
            if text != deanon_text:
                logging.warning(f"Non bidirectional anonymization:\n{text}\n---\n{deanon_text}")

        return anon_text

    def deanonymize(self, text: str, codec: EntityTypeCodec) -> str:
        deanon_text = codec.decode(text)
        return deanon_text


def get_sanitized_input_handling_guide(codec: EntityTypeCodec | None = None) -> str:
    """Codec is used to infer the sanitation replacement format the model may encounter"""
    codec = codec or EntityTypeCodec()
    ip_address = codec.format_indexed_entity_type("IP_ADDRESS", 0)
    credit_card = codec.format_indexed_entity_type("CREDIT_CARD", 2)
    guide = inspect.cleandoc(
        f"""
        Some entities in the tool output and messages may be masked due to privacy or security concerns.
        For example, an IP address may be encoded as {ip_address}, and a credit card number as {credit_card}.
        Same encoding corresponds to same entities. You can reference the encodings in your messages, and 
        they will be unmasked for the user to see.
        """  # noqa: E501
    )
    return guide


def call_privacy_guard_api(
    text: str,
    project_key: str,
    url: str = "https://dssapi-data.kapitalbank.az/masker-on-test/public/api/v1/privacy_guard/PI_masker/run",
    timeout: float = 30.0,
    **payload,
) -> httpx.Response:
    """
    Call the Dataiku masking API endpoint.
    https://dssdesign-data.kapitalbank.az/projects/APILIB/api-designer/privacy_guard/endpoints/
    """
    payload |= {
        "text": text,
        "project_key": project_key,
    }
    response = httpx.post(url, timeout=timeout, json=payload)
    return response



import enum
import functools
import os

import pydantic
import sqlalchemy as sa

from ai_app.utils import PydanticForbidExtra


class Driver(enum.StrEnum):
    oracle = enum.auto()
    postgresql = enum.auto()
    mssql = enum.auto()

    @property
    def name_and_package(self):
        match self:
            case Driver.oracle:
                return "oracle+oracledb"
            case Driver.postgresql:
                return "postgresql+psycopg"
            case Driver.mssql:
                return "mssql+pymssql"

        assert False


@functools.cache
def init_oracle_thick_mode():
    if path := os.getenv("ORACLE_INSTANT_CLIENT_LIB"):
        import oracledb

        oracledb.init_oracle_client(path)


class Connection(PydanticForbidExtra):
    driver: Driver
    host: str
    port: int
    database: str | None = None
    username: str
    password: str
    parameters: dict[str, str] = {}

    @property
    def url(self) -> sa.URL:
        url = sa.URL.create(
            drivername=self.driver.name_and_package,
            host=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
            database=self.database,
            query=self.parameters,
        )
        return url

    def build_engine(self, **kwargs) -> sa.Engine:
        # if self.driver == Driver.oracle:
        #     init_oracle_thick_mode()

        engine = sa.create_engine(self.url, **kwargs)
        return engine


class ConnectionContext(PydanticForbidExtra):
    connection_name: str
    schema_tables: list[str] = []

    @pydantic.model_validator(mode="after")
    def _validate(cls, self):
        self.get_tables_by_schema()
        return self

    def get_tables_by_schema(self) -> dict[str | None, list[str]]:
        tables_by_schema = {}
        for table in self.schema_tables:
            if "." in table:
                schema, table = table.split(".")
            else:
                schema = None

            tables_by_schema.setdefault(schema, []).append(table)

        return tables_by_schema



import httpx

from ai_app.utils import build_url_with_parameters, join_url


def get_openid_metadata_url(base_url: str) -> str:
    """Common OpenID metadata URL for auth providers."""
    url = join_url(base_url, "/.well-known/openid-configuration")
    return url


def get_keycloak_realm_url(base_url: str, realm: str) -> str:
    """Keycloak realm URL."""
    url = join_url(base_url, "realms", realm)
    return url


def to_absolute_route(route: str) -> str:
    route = "/" + route.strip("/")
    return route


def get_minimal_openid_scope() -> str:
    return "email openid profile"


def get_openid_metadata(base_url: str, **kwargs) -> dict:
    response = httpx.get(get_openid_metadata_url(base_url), **kwargs)
    response.raise_for_status()
    metadata = response.json()
    return metadata


def build_logout_redirect_url(
    end_session_endpoint: str, post_logout_redirect_uri: str, id_token: str | None = None
) -> str:
    logout_params = {
        "post_logout_redirect_uri": post_logout_redirect_uri,
    }
    # id_token_hint is recommended by OIDC spec for logout
    if id_token:
        logout_params["id_token_hint"] = id_token

    logout_url = build_url_with_parameters(end_session_endpoint, **logout_params)
    return logout_url



from typing import Iterable

import langchain_openai
import langchain_postgres
import langchain_postgres.v2.indexes
import sqlalchemy as sa

from ai_app.sql import create_index


def build_postgres_vector_store(
    engine: sa.Engine,
    openai_api_key: str,
    table_name: str,
    metadata_columns: Iterable[langchain_postgres.Column] | None = None,
    ensure_metadata_columns_index_is_unique: bool = False,
    retrieved_document_count: int = 5,
    embedding_model: str = "text-embedding-3-small",
    vector_dimension: int = 1024,
    overwrite_existing_table: bool = False,
) -> langchain_postgres.PGVectorStore:
    pg_engine = langchain_postgres.PGEngine.from_connection_string(engine.url)
    embeddings = langchain_openai.OpenAIEmbeddings(
        model=embedding_model,
        dimensions=vector_dimension,
        api_key=openai_api_key,
    )
    vector_index = None
    metadata_columns_names = [c.name for c in metadata_columns]
    if overwrite_existing_table or not sa.inspect(engine).has_table(table_name):
        pg_engine.init_vectorstore_table(
            table_name=table_name,
            vector_size=vector_dimension,
            metadata_columns=metadata_columns,
            overwrite_existing=overwrite_existing_table,
        )
        vector_index = langchain_postgres.v2.indexes.HNSWIndex()
        if metadata_columns_names:
            create_index(
                engine=engine,
                table_name=table_name,
                columns=metadata_columns_names,
                unique=ensure_metadata_columns_index_is_unique,
            )

    store = langchain_postgres.PGVectorStore.create_sync(
        engine=pg_engine,
        table_name=table_name,
        embedding_service=embeddings,
        metadata_columns=metadata_columns_names,
        k=retrieved_document_count,
        fetch_k=5 * retrieved_document_count,
    )
    if vector_index:
        store.apply_vector_index(vector_index)
    else:
        store.reindex()

    return store



import dataclasses
import tempfile
from typing import Iterable

import langchain_chroma
import langchain_community
import langchain_core
import langchain_huggingface
import langchain_text_splitters
import openai
import scipy
import zyphra

from ai_app.config import get_config
from ai_app.utils import wrap_with_xml_tag


def convert_16_bit_wav_to_float(array):
    array = array / 2**15
    return array


@dataclasses.dataclass
class Audio:
    rate: int
    data: Iterable

    def to_tuple(self) -> tuple[int, Iterable[float]]:
        audio = self.rate, self.data
        return audio

    def to_whisper_transformer_format(self) -> dict:
        audio = {"sampling_rate": self.rate, "array": self.data}
        return audio

    def convert_16_bit_wav_to_float(self):
        data = convert_16_bit_wav_to_float(self.data)
        audio = Audio(rate=self.rate, data=data)
        return audio

    def write_wav_to_file(self, file):
        scipy.io.wavfile.write(file, rate=self.rate, data=self.data)

    @classmethod
    def from_wav_file(self, file):
        rate, data = scipy.io.wavfile.read(file)
        audio = Audio(rate=rate, data=data)
        return audio


class SpeechToTextOpenAI:
    def __init__(self):
        self.openai_client = openai.OpenAI(api_key=get_config().get_provider_api_key("openai"))

    def from_audio_sample(
        self,
        audio: Audio,
        model: str = "whisper-1",
        prompt: str | None = None,
        temperature: float = 0.5,
        language: str | None = None,
        **kwargs,
    ) -> str:
        # OpenAI TTS API expects the file to have a name, as it infers the audio format from the filename suffix,
        # and I haven't found a way to assign a name to io.BytesIO or any non-filesystem object.
        temp_file = tempfile.NamedTemporaryFile(suffix=".wav")
        audio.write_wav_to_file(temp_file)
        temp_file.seek(0)
        transcription = self.openai_client.audio.transcriptions.create(
            file=temp_file.file,
            language=language,
            model=model,
            prompt=prompt,
            temperature=temperature,
            **kwargs,
        )
        text = transcription.text
        return text


class TextToSpeechOpenAI:
    def __init__(self):
        self.client = openai.OpenAI(api_key=get_config().get_provider_api_key("openai"))

    def to_audio_bytes(
        self, text, model: str = "tts-1", voice: str = "alloy", speed: float = 1
    ) -> bytes:
        response = self.client.audio.speech.create(
            input=text,
            model=model,
            voice=voice,
            speed=speed,
            response_format="wav",
        )
        audio_bytes = response.read()
        return audio_bytes


class TextToSpeechZyphra:
    """
    Zyphra has an open source Zonos model, which is available in the zonos package https://github.com/Zyphra/Zonos,
    but I couldn't install it. They also have an API with free tier: https://playground.zyphra.com/
    and a client package: https://pypi.org/project/zyphra/.
    """

    def __init__(self):
        self.api_key = ...  # get_bitwarden_secret("c28a662a-f72f-4bb2-afa1-b28b00a23b2d")

    def to_audio_bytes(self, text, speaking_rate: float = 15.0) -> bytes:
        with zyphra.ZyphraClient(api_key=self.api_key) as client:
            audio_bytes = client.audio.speech.create(
                text=text,
                speaking_rate=speaking_rate,
            )

        return audio_bytes


class Retriever:
    """
    https://python.langchain.com/docs/tutorials/qa_chat_history/
    https://langchain-ai.github.io/langgraph/tutorials/rag/langgraph_agentic_rag/
    """

    def __init__(self, directory: str = "data/chroma", document_count_to_retrieve: int = 3):
        embeddings = langchain_huggingface.HuggingFaceEmbeddings(
            model_name="sentence-transformers/all-mpnet-base-v2"
        )
        self.vector_store = langchain_chroma.Chroma(
            embedding_function=embeddings,
            persist_directory=directory,
        )
        self.document_count_to_retrieve = document_count_to_retrieve

        @langchain_core.tools.tool()
        def retrieve(query: str):
            """Retrieve information related to a query."""
            return self.retrieve(query)

        # Todo: be able to directly decorate class methods without the "self" argument being added to function signature.
        self.retrieve_tool = retrieve

    def prepare_data(self):
        loader = langchain_community.document_loaders.WebBaseLoader(
            web_paths=[
                "https://www.kapitalbank.az/en/cards",
                "https://www.kapitalbank.az/en/cards/TaksitCards",
                "https://www.kapitalbank.az/en/cards/simpleCards",
                "https://www.kapitalbank.az/en/cards/GiftCards",
                "https://www.kapitalbank.az/en/cards/studentCard",
                "https://www.kapitalbank.az/en/loans/mikro-biznes-kreditleri",
                # In Azerbaijani:
                # "https://birbank.az",
                # "https://birbank.az/cards/all",
                # "https://birbank.business/en",
                # "https://birbank.business/hesab-xidmetleri",
                # "https://birbank.business/kartlar",
                # "https://birbank.business/kreditler",
                # "https://birbank.business/senedli-emeliyyatlar",
                # "https://birbank.business/elektron-ticaret",
                # "https://birbank.business/pos-xidmetleri",
            ],
        )
        documents = loader.load()
        text_splitter = langchain_text_splitters.RecursiveCharacterTextSplitter(
            chunk_size=1000, chunk_overlap=200
        )
        document_splits = text_splitter.split_documents(documents)
        self.vector_store.add_documents(documents=document_splits)

    def retrieve(self, query: str):
        documents = self.vector_store.similarity_search(query, k=self.document_count_to_retrieve)
        documents = "\n".join(wrap_with_xml_tag("document", d.page_content) for d in documents)
        return documents



import logging
import time
import urllib
import webbrowser
from typing import Iterable

import authlib.integrations.httpx_client
import authlib.integrations.starlette_client
import fastapi
import gradio as gr
import starlette.middleware.sessions
import starlette.responses

from ai_app.auth.utils import (
    build_logout_redirect_url,
    get_minimal_openid_scope,
    get_openid_metadata,
    get_openid_metadata_url,
    to_absolute_route,
)
from ai_app.utils import build_url_with_parameters, compress, decompress, filter_dict_by_keys


class OAuth:
    oauth_registry: authlib.integrations.starlette_client.OAuth
    starlette_client: authlib.integrations.starlette_client.StarletteOAuth2App
    httpx_client: authlib.integrations.httpx_client.OAuth2Client
    openid_metadata: dict[str, str]

    def __init__(
        self,
        provider_base_url: str,
        client_id: str,
        client_secret: str,
        scope: str | None = None,
        oauth: authlib.integrations.starlette_client.OAuth | None = None,
        provider_name: str = "oauth_provider",
        redirect_uri: str = "http://localhost/auth",
    ):
        scope = scope or get_minimal_openid_scope()
        self.oauth_registry = oauth or authlib.integrations.starlette_client.OAuth()
        self.oauth_registry.register(
            name=provider_name,
            client_id=client_id,
            client_secret=client_secret,
            server_metadata_url=get_openid_metadata_url(provider_base_url),
            client_kwargs={"scope": scope, "verify": False},
        )

        self.starlette_client = self.oauth_registry.create_client(provider_name)
        self.httpx_client = authlib.integrations.httpx_client.AsyncOAuth2Client(
            client_id=client_id,
            client_secret=client_secret,
            scope=scope,
            redirect_uri=redirect_uri,
        )
        self.openid_metadata = get_openid_metadata(provider_base_url)

    def interactive_authorization(self) -> tuple[dict, dict]:
        """
        Manually open the authorization URL in the browser, login, and copy the resulting
        redirected URL into the parse_redirected_url method.
        Meant only for testing in interactive environments.
        """
        authorization_url, auth_state = self.httpx_client.create_authorization_url(
            self.openid_metadata["authorization_endpoint"]
        )
        webbrowser.open(authorization_url)
        redirected_url = input("Paste the full redirect URL here: ")
        parsed_url = urllib.parse.urlparse(redirected_url)
        query_params = urllib.parse.parse_qs(parsed_url.query)
        authorization_code = query_params.get("code", [None])[0]
        returned_state = query_params.get("state", [None])[0]
        if auth_state != returned_state:
            raise ValueError("State mismatch. Possible CSRF attack.")

        token = self.httpx_client.fetch_token(
            self.openid_metadata["token_endpoint"],
            code=authorization_code,
        )

        response = self.httpx_client.get(self.openid_metadata["userinfo_endpoint"])
        response.raise_for_status()
        userinfo = response.json()
        return token, userinfo


def get_username_if_authenticated(request: fastapi.Request) -> str | None:
    """
    Returns None if user is not authenticated, otherwise returns the user name,
    which had to be manually saved in the session upon user login.
    Can be directly passed to the Gradio mount_gradio_app method as an auth dependency.
    https://www.gradio.app/docs/gradio/mount_gradio_app#param-mount_gradio_app-auth-dependency
    """
    expires_at = request.session.get("access_token_expires_at")
    user_info = request.session.get("user_info")
    if not user_info or expires_at and expires_at < time.time():
        return

    username = user_info.get("name")
    return username


def check_oauth_error_in_url_params(request: gr.Request):
    error = request.query_params.get("error")
    if error:
        description = request.query_params.get("description")
        raise gr.Error(f"OAuth failed with error: '{error}' and description: '{description}'")


def mount_auth_for_gradio_app(
    oauth_client: authlib.integrations.starlette_client.StarletteOAuth2App,
    gradio_app: gr.Blocks,
    login_gradio_app: gr.Blocks | None = None,
    gradio_app_route: str = "gradio",
    login_gradio_app_route: str = "gradio-login",
    auth_redirect_route: str = "auth",
    app: fastapi.FastAPI | None = None,
    session_crypto_key: str | None = None,
    end_session_endpoint: str | None = None,
    userinfo_fields: Iterable[str] = ("employeeID", "preferred_username", "name", "email"),
    max_id_token_size: int = 0,
    login_gradio_app_kwargs: dict | None = None,
    gradio_app_kwargs: dict | None = None,
) -> fastapi.FastAPI:
    """
    The auth_redirect_path should be the one registered by the client with the OAuth provider.
    The max_id_token_size parameter sets the maximum size of the id_token to be saved in the session.
    If it's too big, the whole session cookie will be dropped, disabling login.
    The id_token is only needed for proper logout handling, so it is not crucial to save it if it's too big.
    Usually the session cookie size is about 4KB, so only necessary data should be stored in it.

    If browser sessions are encoded with different crypto keys, they will not be reused,
    so it's better to persist the key between app restarts.
    A secure key can be generated with secrets.token_hex(32).
    """
    # Todo: move id_token to a server-side session storage, like Redis.
    gradio_app_route = to_absolute_route(gradio_app_route)
    login_gradio_app_route = to_absolute_route(login_gradio_app_route)
    auth_redirect_route = to_absolute_route(auth_redirect_route)
    app = app or fastapi.FastAPI()
    if session_crypto_key:
        app.add_middleware(
            starlette.middleware.sessions.SessionMiddleware, secret_key=session_crypto_key
        )

    @app.get("/")
    def root(request: fastapi.Request):
        """Redirects to the main gradio app if the user is authenticated, otherwise to the login gradio app."""
        username = get_username_if_authenticated(request)
        route = gradio_app_route if username else login_gradio_app_route
        redirect = starlette.responses.RedirectResponse(url=route)
        return redirect

    @app.route("/logout")
    async def logout(request: fastapi.Request):
        id_token = request.session.get("id_token")
        request.session.clear()
        if not end_session_endpoint or not id_token:
            url = "/"
        else:
            id_token = decompress(id_token)
            url = build_logout_redirect_url(
                end_session_endpoint=end_session_endpoint,
                post_logout_redirect_uri=request.url_for("root"),
                id_token=id_token,
            )

        redirect = starlette.responses.RedirectResponse(url=url)
        return redirect

    @app.route("/login")
    async def login(request: fastapi.Request):
        """Redirect to provider login page, which will redirect back to the redirect_url after authentication."""
        redirect_url = request.url_for("auth")
        redirect = await oauth_client.authorize_redirect(request, redirect_url)
        return redirect

    @app.route(auth_redirect_route)
    async def auth(request: fastapi.Request):
        """Try to fetch access token from provider. If successful, save userinfo in session to flag that the user is authenticated."""
        print("auth")
        try:
            access_token = await oauth_client.authorize_access_token(request)
        except authlib.integrations.starlette_client.OAuthError as e:
            url = build_url_with_parameters(
                login_gradio_app_route, error=e.error, description=e.description
            )
        else:
            access_token = dict(access_token)
            compressed_id_token = compress(access_token["id_token"])
            if len(compressed_id_token) <= max_id_token_size:
                request.session["id_token"] = compressed_id_token
            else:
                logging.warning("id_token is too large to be saved in session.")

            request.session["access_token_expires_at"] = access_token["expires_at"]
            request.session["user_info"] = filter_dict_by_keys(
                access_token["userinfo"], userinfo_fields
            )
            url = "/"

        redirect = starlette.responses.RedirectResponse(url)
        return redirect

    if not login_gradio_app:
        with gr.Blocks() as login_gradio_app:
            gr.Button("Login", link="/login")
            login_gradio_app.load(check_oauth_error_in_url_params)

    def auth_redirect_dependency(request: fastapi.Request) -> str | None:
        username = get_username_if_authenticated(request)
        if not username:
            raise fastapi.HTTPException(
                status_code=fastapi.status.HTTP_307_TEMPORARY_REDIRECT,
                detail="Not authenticated",
                headers={"Location": str(request.url_for("root"))},
            )

        return username

    app = gr.mount_gradio_app(
        app,
        login_gradio_app,
        login_gradio_app_route,
        **(login_gradio_app_kwargs or {}),
    )
    app = gr.mount_gradio_app(
        app,
        gradio_app,
        gradio_app_route,
        auth_dependency=auth_redirect_dependency,
        **(gradio_app_kwargs or {}),
    )
    return app



import fastapi
import limits.storage
import limits.strategies
import pydantic

from ai_app.auth import get_username_if_authenticated
from ai_app.utils import (
    HttpMethod,
    PydanticForbidExtra,
    are_rate_limits_hit,
    get_remote_address,
)


class CorsConfig(PydanticForbidExtra):
    allow_origins: list[str] = []
    allow_credentials: bool = False
    allow_methods: list[str] = ["GET", "POST", "PUT", "DELETE", "OPTIONS"]
    allow_headers: list[str] = ["*"]
    allow_origin_regex: str | None = None


def get_user_key_for_rate_limits(request: fastapi.Request) -> str:
    key = get_username_if_authenticated(request) or get_remote_address(request)
    return key


class EndpointRateLimitConfig(PydanticForbidExtra):
    method: HttpMethod
    path: str
    limit: str

    @pydantic.model_validator(mode="after")
    def _validate(cls, self):
        """Check that limit field conforms to limits library format."""
        self.rate_limits
        return self

    @property
    def rate_limits(self) -> list[limits.RateLimitItem]:
        return limits.parse_many(self.limit)

    def should_apply_limit(self, request: fastapi.Request) -> bool:
        return request.method == self.method and self.path in request.url.path

    def get_rate_limit_key(self, request: fastapi.Request) -> tuple[str, ...]:
        key = (self.method.value, self.path, get_user_key_for_rate_limits(request))
        return key


class RateLimitConfig(PydanticForbidExtra):
    storage_uri: str = "memory://"
    endpoint_limits: list[EndpointRateLimitConfig] = []

    def are_rate_limits_hit(
        self, request: fastapi.Request, limiter: limits.strategies.RateLimiter
    ) -> bool:
        limits_with_keys = []
        for endpoint_limit in self.endpoint_limits:
            if endpoint_limit.should_apply_limit(request):
                key = endpoint_limit.get_rate_limit_key(request)
                for rate_limit in endpoint_limit.rate_limits:
                    limits_with_keys.append((rate_limit, key))

        hit = are_rate_limits_hit(limiter, limits_with_keys)
        return hit

    def build_limiter(self) -> limits.strategies.RateLimiter:
        storage = limits.storage.storage_from_string(self.storage_uri)
        limiter = limits.strategies.MovingWindowRateLimiter(storage)
        return limiter



import gradio as gr

from ai_app.apps.sql_agent.backend import Bot
from ai_app.config import get_config
from ai_app.core import BaseApp
from ai_app.frontend import build_model_choice_dropdown


class App(BaseApp):
    name = "SQL agent"
    requires_auth = True

    def __init__(self):
        self.bot = Bot()
        self.respond = self.bot.respond

    def build_gradio_blocks(self, model_choice: gr.Dropdown | None = None) -> gr.Blocks:
        with gr.Blocks(fill_height=True) as app:
            # SQL queries are hard, need a powerful model.
            model_choice = model_choice or build_model_choice_dropdown(models=["gpt-4.1"])
            connection_context_choice = gr.Dropdown(
                choices=sorted(get_config().sql_connection_contexts),
                label="Connection",
            )
            self.plot = gr.Plot(render=False)
            self.build_gradio_chat_interface(
                model_choice=model_choice,
                additional_inputs=[connection_context_choice],
                additional_outputs=[self.plot],
                description=(
                    "The chatbot will help you with SQL queries. Note that due to difficulty of the task, "
                    "it is recommended to use a powerful model."
                ),
                examples=[  # TODO: set connection context for each example.
                    [
                        "What kind of tables do you see?",
                    ],
                    [
                        "Provide details about which structures are struggling with budget, and which ones have a surplus."
                    ],
                    [
                        "Describe average employee tenure by tribe.",
                    ],
                    [
                        "Display a chart based on tribe maturity over time.",
                    ],
                ],
            )
            self.plot.render()

        return app



import datetime
import functools
import importlib
import logging

import gradio as gr

from ai_app.auth import check_oauth_error_in_url_params
from ai_app.config import get_config
from ai_app.core import BaseApp, State
from ai_app.frontend import build_model_choice_dropdown, get_generating_text_javascript
from ai_app.utils import Timer, get_pydantic_model_fields


@functools.cache
def try_build_module_app(module_name: str) -> BaseApp | None:
    try:
        app_module = importlib.import_module(f"ai_app.apps.{module_name}")
    except ImportError:
        logging.error(
            f"Failed to import {module_name} module, probably because of missing optional packages",
            exc_info=True,
        )
        return

    with Timer(f"Initializing app '{app_module.App.name}'"):
        app = app_module.App()

    return app


def build_tab_app(module_name: str, assume_auth: bool, model_choice: gr.Dropdown):
    app = try_build_module_app(module_name)
    if not app or app.requires_auth and not assume_auth:
        return

    # For some reason, the scale argument needs to be passed to make tab fill height.
    with gr.Tab(app.name, scale=1):
        with Timer(f"Building app '{app.name}'"):
            app.build_gradio_blocks(model_choice=model_choice)


def build_tabs(assume_auth: bool, model_choice: gr.Dropdown):
    # Todo: maybe make dynamic at runtime based on "show" values in config.
    for module_name, app_config in get_pydantic_model_fields(get_config().apps).items():
        if app_config.show:
            build_tab_app(
                module_name=module_name,
                assume_auth=assume_auth,
                model_choice=model_choice,
            )


def display_username(request: gr.Request) -> str | None:
    try:
        user_info = State.from_request(request)
    except AssertionError:
        # Catch AssertionError: SessionMiddleware must be installed to access request.session
        return

    if not user_info.name:
        return

    return f"Welcome, {user_info.name}"


def build_page(
    assume_auth: bool | None = None,
    analytics_enabled: bool = False,
    delete_cache_frequency: datetime.timedelta = datetime.timedelta(hours=1),
    delete_cache_age: datetime.timedelta = datetime.timedelta(days=1),
    fill_height: bool = True,
) -> gr.Blocks:
    with gr.Blocks(
        analytics_enabled=analytics_enabled,
        delete_cache=(
            int(delete_cache_frequency.total_seconds()),
            int(delete_cache_age.total_seconds()),
        ),
        fill_height=fill_height,
        js=get_generating_text_javascript(),
    ) as gradio_app:
        with gr.Row():
            model_choice = build_model_choice_dropdown(models=get_config().model_choices)
            if assume_auth is not None:
                markdown = gr.Markdown(rtl=True)
                if assume_auth:
                    gr.Button("Logout", link="/logout")
                    gradio_app.load(display_username, outputs=markdown)
                else:
                    gr.Button("Login", link="/login")
                    gradio_app.load(check_oauth_error_in_url_params)

        build_tabs(assume_auth=bool(assume_auth), model_choice=model_choice)

    return gradio_app


gr.Tab



import functools
import textwrap
from typing import override

import gradio as gr
import langchain_core
import langgraph.prebuilt
import pydantic

from ai_app.config import get_config
from ai_app.core import BaseApp
from ai_app.frontend import build_model_choice_dropdown, get_voice_activity_javascript

from .backend import Audio, Retriever, SpeechToTextOpenAI, TextToSpeechOpenAI


def build_prompt_template() -> langchain_core.prompts.ChatPromptTemplate:
    prompt_template = langchain_core.prompts.ChatPromptTemplate(
        [
            {
                "role": "system",
                "content": "You are a telemarketing salesperson at the bank Birbank.",
            },
            {
                "role": "user",
                "content": textwrap.dedent("""
                    You will call a client of Birbank and suggest him some of Birbank products. 
                    Keep your responses short and enthusiastic. Be proactive and try to end your responses with a question. 
                    Your responses will be read aloud, so do not use markdown formatting, numbered or bullet lists.
                    Respond to the user in the same language he speaks.
                    If the user is interested in some Birbank product, suggest to send him a link with product details in a text message.
                    If appropriate, suggest to schedule a visit to the bank for the user to sign the contract.
                """),
                # """
                # Here is the list of available Birbank cards:
                # - Birbank Star: Our new card with a credit line up to 30,000 manats has everything you need,
                # from simplicity and convenience to cashback and 2x VAT refund.
                # - Birbank Cashback: Birbank card will give you up to 30% cashback, free transfer and withdrawal,
                # double VAT and other benefits.
                # - Birbank Umico: It is a unique card jointly presented by Kapital Bank and Umico, designed for daily shopping,
                # combining credit and installment card opportunities and the opportunity to earn additional Umico bonuses.
                # - Birbank Miles: A unique card that gives a bonus of 1 mile per 1 manat for cashless payments with the card,
                # free air travel, interest-free and commission-free installments.
                # - Birbank Umico Premium: It is a unique card jointly presented by Kapital Bank and Umico, designed for daily shopping,
                # combining credit and installment card opportunities and the opportunity to earn additional Umico bonuses.
                # """,
            },
        ]
    )
    return prompt_template


def get_greeting(language: str | None = None) -> str:
    greetings = {
        "az": "Gününüz xeyir! Birbank kredit və debet kartları haqqında bilmək istərdinizmi?",
        "en": "Good day! Would you like to know about Birbank credit and debit cards?",
    }
    greeting = greetings[language or "en"]
    return greeting


class Response(pydantic.BaseModel):
    """The telemarketer model's response, with optional actions to perform."""

    content: str = pydantic.Field(description="The response that the user will see in the chat.")
    offer_details_to_send_to_client: str | None = pydantic.Field(
        default=None,
        description=(
            "The exact offer or product that the user was interested in and requested details about, to be send to him as a text message, "
            "only if the user explicitly asked for them, or consented to receive."
        ),
    )
    schedule_visit: str | None = pydantic.Field(
        default=None,
        description="The time of a visit to the bank that the user asked for or agreed to schedule.",
    )

    def get_events_list(self) -> list[str]:
        events = []
        if self.offer_details_to_send_to_client:
            events.append(f"Send offer details: {self.offer_details_to_send_to_client}")
        if self.schedule_visit:
            events.append(f"Schedule a visit to the bank: {self.schedule_visit}")

        return events


class App(BaseApp):
    """
    Todo:
    - Provide correct "Birbank" pronunciation to TTS models
    - Pass prompt to TTS model with request for intonation and inflection
    - Try to improve latency by trying different sample rates (librosa.resample, pydub.AudioSegment.export), formats (mp3, webm)
    - Greet user only after he said hello

    Resources:
    - https://huggingface.co/spaces/gradio/omni-mini/blob/main/app.py
    - https://github.com/bklieger-groq/gradio-groq-basics/tree/main/calorie-tracker
    """

    name = "Telemarketer"
    requires_auth = False

    def __init__(
        self, speech_to_text_model=None, text_to_speech_model=None, language: str | None = None
    ):
        self.speech_to_text_model = speech_to_text_model or functools.partial(
            SpeechToTextOpenAI().from_audio_sample, language=language
        )
        self.text_to_speech_model = text_to_speech_model or TextToSpeechOpenAI().to_audio_bytes
        self.language = language
        self.prompt_messages = build_prompt_template().invoke({}).messages

    @property
    @functools.cache
    def retriever(self):
        # Lazy initializing heavy class.
        return Retriever()

    def greet(self):
        greeting = get_greeting(language=self.language)
        greeting = [gr.ChatMessage(role="assistant", content=greeting)]
        return greeting

    def speech_to_text(self, audio: tuple, messages: list[dict]):
        audio = Audio(*audio)
        audio = audio.convert_16_bit_wav_to_float()
        text = self.speech_to_text_model(audio)
        message = gr.ChatMessage(role="user", content=text)
        messages.append(message)
        return messages

    def build_agent(self, model_name: str):
        model_parameters = get_config().build_chat_model_parameters(model_name)
        model = model_parameters.build_model()
        agent = langgraph.prebuilt.create_react_agent(
            model, [self.retriever.retrieve_tool], response_format=Response
        )
        return agent

    def respond(self, model_name: str, messages: list[dict], events: str):
        if messages[-1]["role"] == "assistant":
            return [messages, events]

        agent = self.build_agent(model_name)
        output = agent.invoke({"messages": self.prompt_messages + messages})
        response = output["structured_response"]
        new_events = response.get_events_list()
        events = "\n".join([events] + new_events)
        message = gr.ChatMessage(role="assistant", content=response.content)
        messages.append(message)
        return [messages, events]

    def text_to_speech(self, messages):
        if messages[-1]["role"] == "user":
            return

        text = messages[-1]["content"]
        audio = self.text_to_speech_model(text)
        audio = gr.Audio(audio, autoplay=True, recording=False)
        return audio

    def listen(self):
        audio = gr.Audio(None, autoplay=False, recording=True)
        return audio

    @override
    def build_gradio_blocks(self, model_choice: gr.Dropdown | None = None) -> gr.Blocks:
        with gr.Blocks(js=get_voice_activity_javascript()) as app:
            model_choice = model_choice or build_model_choice_dropdown()
            start = gr.Button("Start", variant="primary")
            chatbot = gr.Chatbot(type="messages")
            audio = gr.Audio(sources="microphone", format="wav")
            events = gr.TextArea(label="Events")

            start.click(self.greet, outputs=chatbot)
            chatbot.change(
                self.respond, inputs=[model_choice, chatbot, events], outputs=[chatbot, events]
            ).then(
                self.text_to_speech,
                inputs=chatbot,
                outputs=audio,
            )
            audio.stop_recording(self.speech_to_text, inputs=[audio, chatbot], outputs=chatbot)
            audio.stop(self.listen, outputs=audio)

        return app



import datetime
import inspect
from typing import override

import gradio as gr
import langchain_core.messages
import polars as pl
import pydantic

from ai_app.ai_utils import try_converting_to_openai_flex_tier
from ai_app.config import get_chat_model, get_config
from ai_app.core import BaseApp
from ai_app.external.google import Drive, Spreadsheets, build_drive, build_spreadsheets
from ai_app.frontend import build_model_choice_dropdown
from ai_app.utils import file_to_base64



class PresentationReview(pydantic.BaseModel):
    review: str = pydantic.Field(
        description=inspect.cleandoc("""
            The presentation review with Markdown formatting in the following schema:
            ```
            **Review:**
            <Succinct review tied to rubric criteria>

            **Score:** <score>

            **Suggestions for Improvement:**
            <List of concrete suggestions>
            ```
    """)
    )
    score: int = pydantic.Field(
        ge=1,
        le=5,
        description=inspect.cleandoc("""
            A score from 1 to 5, based on the presentation criteria.
    """),
    )


def get_prompt():
    prompt = inspect.cleandoc("""
        # Role and Objective
        You are an executive presentation reviewer. Your purpose is to strictly, fairly, and 
        consistently assess presentation quality based on clear, business-relevant criteria.

        # Instructions
        First, review the presentation according to the following criteria:
        - Clarity of objectives and scope
        - Business impact, linkage to strategy
        - Evidence/metrics (baselines, changes, ROI, timeframes)
        - Structure/storytelling (problem → approach → results → decisions)
        - Risks, dependencies, blockers
        - Visual effectiveness, communication quality
        - Description of past month results
        - Next month's plans and objectives

        Then, assign a score to the presentation according to the criteria,
        and follow up with specific, actionable suggestions to improve the presentation.

        # Verbosity
        - Keep total response under 200 words.
        - Do not provide extra explanation or context outside the review format.
    """)
    prompt = langchain_core.messages.SystemMessage(prompt)
    return prompt


class App(BaseApp):
    name = "Review presentation"
    requires_auth = False

    def __init__(self):
        self.spreadsheets: Spreadsheets = build_spreadsheets(
            spreadsheet_id="10Pq9Y43_QrU5tU2YCEK39K-q-3RPZDdIVh3GM-dfxnY",
            service_account_info=get_config().secrets.google_service_account_info.get_secret_value(),
        )
        tribes_df = self.spreadsheets.read_sheet("tribes description")
        self.tribes = sorted(set(tribes_df["Tribe"]) - {"Terminated"})
        self.sheet_name = "OKR presentations"
        self.gdrive_folder_id = "1X3jVXMYyNPk2AJtEI3PDmz2080NEucgM"

    def review(self, filepath: str, tribe: str | None, model: str = "gpt-5"):
        # TODO:
        # - Improve prompt, ensure it aligns with criterias.
        # - Be able to process large pdf in chunks.
        prompt = get_prompt()
        message = langchain_core.messages.HumanMessage(
            content=[
                {
                    "type": "file",
                    "mime_type": "application/pdf",
                    "source_type": "base64",
                    "filename": "presentation.pdf",
                    "data": file_to_base64(filepath),
                },
            ]
        )
        model = get_chat_model(model)
        model = try_converting_to_openai_flex_tier(model)
        model = model.with_structured_output(PresentationReview)
        review = model.invoke([prompt, message])

        gdrive: Drive = build_drive(
            get_config().secrets.google_authorized_user_info.get_secret_value()
        )
        url = gdrive.upload(filepath, parents=[self.gdrive_folder_id])

        data = {
            "Date": str(datetime.date.today()),
            "Tribe": tribe,
            "Score": review.score,
            "Reason": review.review,
            "URL": url,
        }
        df = self.spreadsheets.read_sheet(self.sheet_name)
        df = pl.concat([df, pl.DataFrame(data)], how="diagonal_relaxed")
        self.spreadsheets.write_sheet(self.sheet_name, df)
        return review.review, review.score

    @override
    def build_gradio_blocks(self, model_choice: gr.Dropdown | None = None) -> gr.Blocks:
        with gr.Blocks() as app:
            model_choice = model_choice or build_model_choice_dropdown()
            gr.Markdown(
                inspect.cleandoc("""
                Note that gpt-5 will be used for review, and the model dropdown above is ignored.
            """)
            )
            with gr.Row():
                gr.Markdown()  #
                tribe = gr.Dropdown(self.tribes, label="Tribe", value=None)
                review = gr.UploadButton(
                    "Upload a PDF presentation to review",
                    variant="primary",
                    file_types=[".pdf"],
                )
                gr.Markdown()

            gr.Markdown("---\n## Output")
            with gr.Row():
                score = gr.TextArea(label="Score", lines=1, scale=0)
                feedback = gr.Markdown(label="Feedback")

            review.upload(
                self.review,
                inputs=[review, tribe],
                outputs=[feedback, score],
            )

        return app



import uuid
from typing import override

import bidict
import gradio as gr
import pandas as pd
import pydantic

from ai_app.core import BaseApp
from ai_app.external.atlassian import (
    IssueLink,
    build_issue_url,
    fetch_all_jira_project_keys,
    fetch_all_jira_usernames,
)
from ai_app.frontend import build_model_choice_dropdown

from .backend import JiraTest, JiraTestGenerator, JiraTestStep, TestType


class State(pydantic.BaseModel):
    issue_key: str
    run_id: uuid.UUID


def prettify(col: str) -> str:
    col = col.replace("_", " ")
    col = col.capitalize()
    return col


class App(BaseApp):
    name = "Generate Jira manual test"
    requires_auth = False

    def __init__(self):
        self.jira_test_generator = JiraTestGenerator()
        model_fields = list(JiraTestStep.model_fields)
        self.step_columns = [prettify(i) for i in model_fields]
        self.step_field_to_column = bidict.bidict()
        self.step_field_to_column.putall(zip(model_fields, self.step_columns))

    def generate_test(self, model_name: str, issue_key: str, additional_prompt: str):
        jira_test_with_evaluation, run_id = self.jira_test_generator.generate(
            model_name, issue_key, additional_prompt
        )
        if jira_test_with_evaluation is None:
            raise gr.Error("Model couldn't generate a structured Jira test, try different model.")

        evaluation = jira_test_with_evaluation.evaluation
        jira_test = jira_test_with_evaluation.jira_test
        information_quality = gr.BarPlot(
            pd.DataFrame({"Quality": [evaluation.information_quality], "x": ["placeholder"]}),
            y="Quality",
            y_lim=(0, 5),
        )
        steps = pd.DataFrame(jira_test.model_dump()["steps"])
        steps = steps.rename(columns=self.step_field_to_column)
        outputs = [
            information_quality,
            evaluation.feedback,
            evaluation.improvement_request,
            jira_test.title,
            jira_test.short_description,
            jira_test.type.to_jira_field(),
            steps,
            State(issue_key=issue_key, run_id=run_id),
            gr.Button(interactive=True),
        ]
        return outputs

    def create_jira_test(
        self,
        title: str,
        description: str,
        test_type: str,
        steps: pd.DataFrame,
        state: State,
        project: str,
        assignee: str,
        issue_link: str | None,
    ) -> str:
        test_type = TestType.from_jira_field(test_type)
        steps = steps.rename(columns=self.step_field_to_column.inverse)
        steps = steps.to_dict(orient="records")
        steps = [JiraTestStep(**step) for step in steps]
        steps = [step for step in steps if step.action and step.expected_result]
        jira_test = JiraTest(
            title=title, short_description=description, test_type=test_type, steps=steps
        )
        issue = self.jira_test_generator.create_jira_test(
            jira_test=jira_test, project=project, assignee=assignee
        )
        self.jira_test_generator.langsmith_client.create_feedback(
            run_id=state.run_id, key="is_generated_jira_manual_test_used", value=True
        )
        issue_key = issue["key"]
        if issue_link:
            issue_link = IssueLink(issue_link)
            issue_link.link_issues(self.jira_test_generator.jira, issue_key, state.issue_key)

        issue_url = build_issue_url(self.jira_test_generator.jira.url, issue_key)
        markdown_issue_url = f"### <center>{issue_url}</center>"
        return markdown_issue_url

    @override
    def build_gradio_blocks(self, model_choice: gr.Dropdown | None = None) -> gr.Blocks:
        with gr.Blocks() as app:
            gr.Markdown(
                "This app facilitates the creation of Jira manual tests for existing issues using an AI model. "
                "The model will be provided the attributes of an existing issue that you need to test and the additional prompt that you may provide. "
            )
            model_choice = model_choice or build_model_choice_dropdown()
            with gr.Row():
                issue_key = gr.Text(
                    placeholder="PFM-3452",
                    label="Issue key",
                    info="The Jira issue to be tested",
                    min_width=100,
                )
                additional_prompt = gr.TextArea(
                    placeholder="Please write a negative test",
                    label="Additional prompt",
                    info="Any additional instructions for the model, along with extra information needed to generate a quality manual test",
                    max_length=20_000,
                    scale=8,
                    lines=3,
                )

            generate_test = gr.Button("Generate Jira test", variant="primary")
            with gr.Row(variant="compact"):
                evaluation = [
                    gr.BarPlot(pd.DataFrame(), label="Provided data quality", min_width=120),
                    gr.Text(label="Feedback", scale=5),
                    gr.Text(label="Improvement request", scale=5),
                ]

            gr.Markdown("You can modify the generated fields below directly within this app.")
            jira_outputs = [
                gr.Text(label="Test title", interactive=True),
                gr.TextArea(label="Test description", lines=3, interactive=True),
                gr.Dropdown(
                    label="Test type",
                    choices=[i.to_jira_field() for i in TestType],
                    value=TestType.draft.to_jira_field(),
                    interactive=True,
                ),
                gr.Dataframe(
                    label="Test steps",
                    headers=self.step_columns,
                    col_count=(len(self.step_columns), "fixed"),
                    row_count=(3, "dynamic"),
                    datatype=["markdown"],
                    wrap=True,
                    interactive=True,
                ),
                gr.State(),
            ]

            gr.Markdown(
                f"If you're satisfied with the test quality, you can choose the Jira project and assignee "
                f"and create the test in [Jira]({self.jira_test_generator.jira})."
            )
            with gr.Row():
                # Todo: use TTL cache for Jira based function calls.
                project_keys = fetch_all_jira_project_keys(self.jira_test_generator.jira)
                project = gr.Dropdown(project_keys, label="Jira project for the test")
                usernames = fetch_all_jira_usernames(self.jira_test_generator.jira)
                assignee = gr.Dropdown(usernames, label="Jira assignee for the test")
                issue_link = gr.Dropdown(
                    list(IssueLink) + [None], label="Link from the test to the provided issue"
                )
                with gr.Column():
                    create_test = gr.Button(
                        "Create Jira test", variant="primary", interactive=False
                    )
                    created_test_url = gr.Markdown(height=70)
                    create_test.click(
                        self.create_jira_test,
                        inputs=jira_outputs + [project, assignee, issue_link],
                        outputs=created_test_url,
                    )

            generate_test.click(
                self.generate_test,
                inputs=[model_choice, issue_key, additional_prompt],
                outputs=evaluation + jira_outputs + [create_test],
            )

        return app



import datetime
import textwrap

import pydantic
import sqlalchemy as sa
import sqlmodel

from ai_app.config import get_ai_postgres_engine
from ai_app.external.atlassian import build_request_url
from ai_app.sql import build_foreign_key_constraint, open_session
from ai_app.utils import get_utc_now


class JiraRequestPrimaryKey(sqlmodel.SQLModel):
    service_desk_id: int = sqlmodel.Field(
        primary_key=True,
        alias="Service desk id",
    )
    request_type_id: int = sqlmodel.Field(
        primary_key=True,
        alias="Request type id",
    )

    @property
    def primary_key_tuple(self) -> tuple[int, int]:
        key = self.service_desk_id, self.request_type_id
        return key

    def __hash__(self) -> int:
        return hash(self.primary_key_tuple)

    @classmethod
    def get_sa_primary_key(cls):
        key = sa.tuple_(
            cls.service_desk_id,
            cls.request_type_id,
        )
        return key

    def build_url(self, **parameters) -> str:
        url = build_request_url(
            service_desk_id=self.service_desk_id,
            request_type_id=self.request_type_id,
            **parameters,
        )
        return url


class JiraRequestCommonColumns(JiraRequestPrimaryKey):
    updated_at: datetime.datetime | None = sqlmodel.Field(
        default_factory=get_utc_now,
        sa_column_kwargs={"onupdate": get_utc_now},
    )


def get_index_column_names() -> list[str]:
    columns = ["service_desk_id", "request_type_id"]
    assert all(c in JiraRequestCommonColumns.model_fields for c in columns)
    return columns


def build_row_index(service_desk_id: int, request_type_id: int) -> dict:
    index = dict(service_desk_id=service_desk_id, request_type_id=request_type_id)
    return index


# Note that table=True disables Pydantic validation at instantiation.
class JiraRequest(JiraRequestCommonColumns, table=True):
    """
    Contains Jira request metadata extracted from Jira API.

    Input fields for the request type:
    If present, the field named "Approvers" should contain the names of employees that need to
    approve the request, which are usually the user's superiors or owners of some systems related to
    the request.
    If present, the "Attachment" field allows for uploading files together with the request,
    such as screenshots or text documents.
    """

    name: str = sqlmodel.Field(description="Jira request type name")
    description: str | None = sqlmodel.Field(
        default=None, description="Jira request type description"
    )
    input_request_fields: list["InputRequestField"] = sqlmodel.Relationship()

    def build_markdown_url(self, **parameters) -> str:
        url = self.build_url(**parameters)
        url = f"[{self.name}]({url})"
        return url


class InputRequestField(JiraRequestCommonColumns, table=True):
    """A request field that needs to be provided by the user when creating a Jira request."""

    __table_args__ = (build_foreign_key_constraint(JiraRequest),)

    field_id: str = sqlmodel.Field(primary_key=True)
    name: str
    valid_values: list[str] | None = sqlmodel.Field(
        default=None,
        description="If not null, then the value for this field must be from this list, otherwise it can be any string.",
        nullable=True,
        sa_type=sa.ARRAY(sa.String),
    )
    can_be_passed_as_url_parameter: bool
    description: str | None = None


class AiJiraRequestGuideContent(sqlmodel.SQLModel):
    """
    A guide to help new users understand the appropriate use of the specific Jira request type
    through clear examples and generalized scenarios, avoiding direct quotes from resolved issues
    to protect privacy and enhance learning.
    """

    use_case: str = sqlmodel.Field(
        description=textwrap.dedent("""
            Comprehensive description of when to use this request type and what problems it solves.
            Include common user queries and problem statements to help with RAG-based matching
            in support chats. Focus on real-world scenarios and typical user needs.
        """)
    )
    resolution_pipeline: str = sqlmodel.Field(
        description=(
            "The usual process that takes place in order to resolve the issue, "
            "along with an example."
        )
    )
    self_help: str | None = sqlmodel.Field(
        default=None,
        description="The ways in which the user can try resolving the issue themselves, if any.",
    )


class AiJiraRequestGuide(AiJiraRequestGuideContent, JiraRequestCommonColumns, table=True):
    """Contains AI generated summary and resolution guide for a Jira request type."""

    __table_args__ = (build_foreign_key_constraint(JiraRequest),)

    input_request_fields: list["AiInputRequestFields"] = sqlmodel.Relationship()


class AiInputRequestFields(JiraRequestCommonColumns, table=True):
    __table_args__ = (build_foreign_key_constraint(AiJiraRequestGuide),)

    field_id: str = sqlmodel.Field(primary_key=True)
    description: str | None = None


class ManualJiraRequestGuide(JiraRequestCommonColumns, table=True):
    """
    Contains manual summary, problem description and additional handling instructions for a Jira
    request type.
    """

    __table_args__ = (build_foreign_key_constraint(JiraRequest),)

    do_recommend: bool = sqlmodel.Field(
        default=True,
        alias="Do recommend",
        description="Whether this request type should be recommended to users.",
    )
    use_case: str | None = sqlmodel.Field(
        default=None,
        alias="Use case",
        description="Use case for this request type, mainly used to modify AI generated use case.",
    )
    resolution: str | None = sqlmodel.Field(
        default=None,
        alias="Resolution",
        description=(
            "The usual process that takes place in order to resolve the issue, "
            "mainly used to modify AI generated resolution and self help."
        ),
    )


def get_saved_request_type_primary_keys(
    engine: sa.Engine, *where_clauses
) -> set[JiraRequestPrimaryKey]:
    statement = sqlmodel.select(JiraRequest.service_desk_id, JiraRequest.request_type_id)
    statement = statement.where(*where_clauses)
    with open_session(engine) as session:
        result = session.exec(statement)
        primary_keys = {JiraRequestPrimaryKey(**pk) for pk in result.mappings()}

    return primary_keys


class JiraRequestType(pydantic.BaseModel):
    meta: JiraRequest
    ai: AiJiraRequestGuide
    manual: ManualJiraRequestGuide


def fetch_guides(
    primary_keys: list[JiraRequestPrimaryKey] | None = None,
    only_recommended: bool = False,
    with_input_fields: bool = False,
    engine: sa.Engine | None = None,
) -> list[JiraRequestType]:
    statement = sqlmodel.select(JiraRequest, AiJiraRequestGuide, ManualJiraRequestGuide).where(
        AiJiraRequestGuide.get_sa_primary_key().in_([pk.primary_key_tuple for pk in primary_keys])
        if primary_keys is not None
        else sa.true(),
        AiJiraRequestGuide.get_sa_primary_key() == ManualJiraRequestGuide.get_sa_primary_key(),
        AiJiraRequestGuide.get_sa_primary_key() == JiraRequest.get_sa_primary_key(),
        ManualJiraRequestGuide.do_recommend if only_recommended else sa.true(),
    )
    if with_input_fields:
        statement = statement.options(
            sa.orm.selectinload(JiraRequest.input_request_fields),
            sa.orm.selectinload(AiJiraRequestGuide.input_request_fields),
        )

    with open_session(engine or get_ai_postgres_engine()) as session:
        guides = list(session.exec(statement))

    guides = [JiraRequestType(meta=meta, ai=ai, manual=manual) for meta, ai, manual in guides]
    return guides


def fetch_use_case_guides_for_requests(primary_keys: list[JiraRequestPrimaryKey]) -> list[dict]:
    guides: list[JiraRequestType] = fetch_guides(primary_keys, only_recommended=True)
    guides = [
        dict(
            service_desk_id=guide.meta.service_desk_id,
            request_type_id=guide.meta.request_type_id,
            request_url=guide.meta.build_markdown_url(),
            request_type_name=guide.meta.name,
            do_recommend_to_user=guide.manual.do_recommend,
            use_case_summary=guide.ai.use_case,
            manually_specified_use_case_overrides=guide.manual.use_case,
        )
        for guide in guides
    ]
    return guides



import contextlib
import datetime
import enum
import functools
import inspect
import json
import logging
import os
import re
import textwrap
import warnings
from typing import Any, AsyncGenerator, Callable, Coroutine, Iterable, NamedTuple

import bir_mcp
import langchain.tools
import langchain_core.language_models
import langchain_core.messages.utils
import langchain_core.prompts
import langsmith
import pydantic

from ai_app.external.providers import get_common_models_meta_attributes
from ai_app.frontend import get_generating_text_span
from ai_app.pii import AnonCodec, EntityTypeCodec
from ai_app.utils import wrap_with_xml_tag


def set_langsmith_environment_variables(
    langsmith_api_key: str,
    langsmith_project: str | None = None,
    langsmith_endpoint: str | None = "https://eu.api.smith.langchain.com",
):
    """
    If LangSmith variables are not set, no errors will be thrown,
    but all logs and traces will be ignored.
    """
    logging.info("Setting LangSmith environment variables")
    os.environ["LANGSMITH_TRACING"] = "true"
    os.environ["LANGSMITH_API_KEY"] = langsmith_api_key
    if langsmith_project:
        os.environ["LANGSMITH_PROJECT"] = langsmith_project
    if langsmith_endpoint:
        os.environ["LANGSMITH_ENDPOINT"] = langsmith_endpoint
        os.environ["LANGCHAIN_ENDPOINT"] = langsmith_endpoint


def fetch_root_langsmith_run() -> langsmith.RunTree | None:
    client = langsmith.client.Client()
    run = langsmith.get_current_run_tree()
    if not run:
        return

    root_run = client.read_run(run.trace_id)
    return root_run


class StructuredOutputMethod(enum.StrEnum):
    json_mode = enum.auto()
    json_schema = enum.auto()
    function_calling = enum.auto()


def prepare_llm_input(text: str | None, max_length: int | None = None) -> str:
    """
    Main purpose of this function is to reduce the length of the input to the LLM, while keeping all
    the information.
    """
    if not text:
        return ""

    text = re.sub(" +", " ", text)
    text = re.sub(r"\n+", r"\n", text)
    if max_length:
        text = text[:max_length]

    return text


def invoke_model(
    model,
    prompt_template: langchain_core.prompts.ChatPromptTemplate,
    prompt_params: dict,
    trace: langsmith.trace | None = None,
    max_prompt_param_length: int | None = None,
) -> tuple[Any, str | None]:
    prompt_params = {
        k: prepare_llm_input(v, max_length=max_prompt_param_length)
        for k, v in prompt_params.items()
    }
    with trace if trace else contextlib.nullcontext() as run:
        prompt = prompt_template.invoke(prompt_params)
        model_output = model.invoke(prompt)
        run_id = run.id if run else None

    return model_output, run_id


def to_formatted_markdown(text: str, kind: str = "json") -> str:
    text = f"```{kind}\n{text}\n```"
    return text


def to_json_llm_input(value: dict | list, indent: int | None = 0) -> str:
    string = json_dumps_for_ai(value, indent=indent)
    string = to_formatted_markdown(string)
    return string


def get_pydantic_config_dict_for_openai_structured_output() -> pydantic.ConfigDict:
    """
    https://platform.openai.com/docs/guides/structured-outputs?api-mode=responses#supported-schemas
    There may be ready utilities to transform Pydantic model to OpenAI structured output JSON
    schema, but I haven't found one.
    Another option is to override schema_generator in model_json_schema, but the documentation is
    lacking.
    https://docs.pydantic.dev/latest/api/json_schema/#pydantic.json_schema.GenerateJsonSchema
    """

    def schema_extra(schema: dict, model) -> None:
        for property in schema.get("properties", {}).values():
            property.pop("title", None)
            property.pop("default", None)

    config_dict = pydantic.ConfigDict(
        extra="forbid",
        json_schema_extra=schema_extra,
        json_schema_serialization_defaults_required=False,
    )
    return config_dict


class SummarizationStateBase(pydantic.BaseModel):
    items_summarized: int = pydantic.Field(
        default=0,
        description=inspect.cleandoc("""
            Number of items summarized, a meta field which automatically updates. 
            Should not be provided by the LLM.
        """),
    )


class Summarizer:
    def __init__(self, model, max_tokens: int, context: str | None = None):
        self.model = model
        self.max_tokens = max_tokens
        self.context = (
            "## Summarization context:\n"
            + wrap_with_xml_tag("summarization_context", context, with_new_lines=True)
            if context
            else ""
        )

    def build_prompt(
        self, state: SummarizationStateBase | type[SummarizationStateBase], items: list[str]
    ) -> str:
        current_state = (
            ""
            if isinstance(state, type)
            else f"## Current summarization state:\n{to_json_llm_input(state.model_dump())}\n"
        )
        prompt = (
            textwrap.dedent("""
            ## Task
            Update the summarization state by incorporating information from new items while 
            maintaining consistency with previous data and context.

            ## Instructions
            1. Carefully analyze the current summarization state, context (if any) and new items
            2. Identify key information, patterns, and insights from the new items
            3. Update the state by:
                - Adding new information
                - Refining existing information when new evidence contradicts or enhaces it
                - Maintaining consistency with previous data
                - Take into account the number of items already summarized by the existing state to 
                    estimate how accurate the provided state is, the more items seen, the more 
                    stable and reliable the current state is
            4. Ensure the updated state is:
                - Comprehensive: Covers all important aspects
                - Concise: Avoids redundancy
                - Accurate: Based strictly on the provided data
                - Well-structured: Follows the required output format
            """)
            + f"{current_state}{self.context}\nNew items:\n{to_json_llm_input(items)}\n"
        )
        return prompt

    def update_summarization_state(
        self,
        state: SummarizationStateBase | type[SummarizationStateBase],
        items: list[str],
    ) -> SummarizationStateBase | type[SummarizationStateBase]:
        if not items:
            return state

        if isinstance(state, type):
            state_type = state
            items_summarized = 0
        else:
            state_type = type(state)
            items_summarized = state.items_summarized

        prompt = self.build_prompt(state, items)
        model = self.model.with_structured_output(state_type)
        new_state = model.invoke(prompt)
        new_state.items_summarized = items_summarized + len(items)
        return new_state

    def summarize(
        self, state: SummarizationStateBase | type[SummarizationStateBase], items: list[str]
    ) -> SummarizationStateBase | None:
        """Returns None if no items were provided, otherwise an updated summarization state."""
        item_buffer = []
        for item in items:
            if len(item) > self.max_tokens:
                raise RuntimeError("Item is too large to be summarized.")

            if sum(len(i) for i in item_buffer + [item]) > self.max_tokens:
                state = self.update_summarization_state(state, item_buffer)
                item_buffer = []

            item_buffer.append(item)

        state = self.update_summarization_state(state, item_buffer)
        state = None if isinstance(state, type) else state
        return state


def extract_tool_calls_and_messages(
    messages: list[langchain_core.messages.BaseMessage],
    tool_names: Iterable[str] | None = None,
    only_success: bool = True,
) -> tuple[list[langchain_core.messages.ToolCall], list[langchain_core.messages.ToolMessage]]:
    ai_tool_calls = []
    tool_messages = []
    for message in messages:
        match message:
            case langchain_core.messages.AIMessage(tool_calls=tool_calls):
                ai_tool_calls.extend(tool_calls)
            case langchain_core.messages.ToolMessage(name=name, status=status):
                if only_success and status != "success":
                    continue
                if tool_names and name not in tool_names:
                    continue

                tool_message = langchain_core.messages.ToolMessage(
                    **message.model_dump(include=["content", "name", "tool_call_id"])
                )
                tool_messages.append(tool_message)

    tool_call_ids = {message.tool_call_id for message in tool_messages}
    ai_tool_calls = [call for call in ai_tool_calls if call["id"] in tool_call_ids]
    return ai_tool_calls, tool_messages


def process_nested_basic_datatypes(
    value: str | dict | list | int | None,
    process_string: Callable[[str], str],
):
    """Can be used to recursively process output of a tool, looking for string values."""

    def recursive_process(value):
        match value:
            case str():
                value = process_string(value)
            case dict():
                value = {k: recursive_process(v) for k, v in value.items()}
            case list():
                value = [recursive_process(v) for v in value]

        return value

    value = recursive_process(value)
    return value


def json_dumps_for_ai(value, **kwargs) -> str:
    if not isinstance(value, str):
        value = json.dumps(value, ensure_ascii=False, **kwargs)

    return value


def wrap_with_output_processor(
    function: Callable | Coroutine,
    process_full_output: Callable | None = None,
    process_string: Callable[[str], str] | None = None,
):
    """
    Wrap the function or coroutine to process its ouput before it is passed to the model.
    Useful for model input sanitation or truncation.
    The output of tool will be converted to string.
    """

    def process_output(result) -> str:
        if process_full_output:
            result = process_full_output(result)
        elif process_string:
            result = process_nested_basic_datatypes(result, process_string=process_string)

        result = json_dumps_for_ai(result)
        return result

    @functools.wraps(function)
    async def awrapped(*args, **kwargs):
        result = await function(*args, **kwargs)
        result = process_output(result)
        return result

    @functools.wraps(function)
    def wrapped(*args, **kwargs):
        result = function(*args, **kwargs)
        result = process_output(result)
        return result

    wrap = awrapped if inspect.iscoroutinefunction(function) else wrapped
    return wrap


def render_tool_call(name: str, args: dict, response: str | None = None) -> str:
    args = to_formatted_markdown(json.dumps(args, indent=2, ensure_ascii=False))
    if not response:
        status = "Calling tool"
        response = ""
    else:
        status = "Tool call"
        try:
            response = json.loads(response)
        except json.JSONDecodeError:
            response = to_formatted_markdown(response, kind="text")
        else:
            response = to_formatted_markdown(json.dumps(response, indent=2, ensure_ascii=False))

        response = inspect.cleandoc("""
            **Response:**

            {response}
        """).format(response=response)

    html = textwrap.dedent(
        """
        <details>
        <summary><b>{status} <code>{name}</code></b></summary>
        
        **Arguments:**
        {args}
        {response}
        </details>
        
        """
    ).format(
        status=status,
        name=name,
        args=args,
        response=response,
    )
    return html


class AstreamResponse(NamedTuple):
    formatted_response: str
    state: dict


async def astream_response_messages(
    agent,
    messages: list[langchain_core.messages.BaseMessage],
    stream_ai_tokens: bool = False,
    with_spinner: bool = True,
) -> AsyncGenerator[AstreamResponse, None]:
    """
    Be careful when processsing payloads from different stream modes - depending on
    steam mode generation order, an "update" may be emitted before "values", thus
    leading to inconsistent state, if it depends on both stream modes.
    """
    state = {}
    new_messages: list[langchain_core.messages.BaseMessage] = []
    tool_responses = {}
    formatted_response = ""
    async for stream_mode, payload in agent.astream(
        {"messages": messages},
        stream_mode=["values", "messages", "updates"],
    ):
        match stream_mode:
            case "values":
                state = payload
                continue

            case "messages":
                message, metadata = payload
                if isinstance(message, langchain_core.messages.AIMessageChunk) and message.content:
                    new_messages.append(message)
                else:
                    continue

            case "updates":
                for update_value in payload.values():
                    for message in update_value.get("messages", []):
                        new_messages.append(message)
                        # Cover cases for custom agents, which do not always emit "tools" updates.
                        if isinstance(message, langchain_core.messages.ToolMessage):
                            tool_responses[message.tool_call_id] = message.content

                # Optimization for "tools" node.
                if tools_update := payload.get("tools"):
                    for tool_message in tools_update["messages"]:
                        tool_responses[tool_message.tool_call_id] = tool_message.content

        formatted_messages = []
        for message in new_messages:
            match message:
                case langchain_core.messages.AIMessageChunk() if stream_ai_tokens:
                    formatted_messages.append(message.content)

                case langchain_core.messages.AIMessage():
                    if not stream_ai_tokens:
                        formatted_messages.append(message.content)

                    for tool_call in message.tool_calls:
                        rendered_message = render_tool_call(
                            name=tool_call["name"],
                            args=tool_call["args"],
                            response=tool_responses.get(tool_call["id"]),
                        )
                        formatted_messages.append(rendered_message)

        formatted_response = "".join(formatted_messages)
        yield AstreamResponse(
            formatted_response=formatted_response
            # Do not add spinner if generating AI message chunks.
            + (get_generating_text_span() if with_spinner and stream_mode != "messages" else ""),
            state=state,
        )

    yield AstreamResponse(formatted_response=formatted_response, state=state)


def trim_history_messages(
    messages,
    max_tool_content_len: int = 10_000,
    tool_truncation_placeholder: str = (
        "\n[CONTENT END] The tool output was truncated to reduce the conversation history size. "
        "You can call the tool again to get the full output."
    ),
):
    """
    Should not be used as the pre_model_hook, as then the model will never see the full tool output.
    The tool message trimming is rather straightforward and safe way to reduce the history size.
    For comparison, if deciding to remove tool messages, it is also necessary to remove the tool
    calls so as not to confuse the model and make it think that the tool calls were unsuccessful.
    But when removing tool messages is needed, it is possible to use the
    langchain_core.messages.filter_messages function.
    """
    trimmed_messages = []
    for message in messages:
        match message:
            case langchain_core.messages.ToolMessage():
                message.content = bir_mcp.utils.truncate_text(
                    text=message.content,
                    max_length=max_tool_content_len,
                    placeholder=tool_truncation_placeholder,
                )

        trimmed_messages.append(message)

    return trimmed_messages


class PreModelHook:
    def __init__(
        self,
        model: langchain_core.language_models.BaseChatModel | None = None,
        max_tokens: int | None = None,
    ):
        if model:
            token_counter = model.get_num_tokens_from_messages
            try:
                token_counter([langchain_core.messages.HumanMessage("test")])
            except NotImplementedError:
                token_counter = langchain_core.messages.utils.count_tokens_approximately

        if not max_tokens:
            common_models = get_common_models_meta_attributes()
            if model and model.model_name in common_models:
                max_tokens = common_models[model.model_name].context_window
            else:
                max_tokens = 200_000

        self.token_counter = token_counter
        self.max_tokens = max_tokens

    def __call__(self, state: dict) -> dict:
        trimmed_messages = langchain_core.messages.trim_messages(
            state["messages"],
            token_counter=self.token_counter,
            max_tokens=self.max_tokens,
            start_on="human",
            include_system=True,
            allow_partial=False,
        )
        return {"llm_input_messages": trimmed_messages}


def anonymize_tool_output(
    tool: langchain.tools.StructuredTool, anon_codec: AnonCodec, codec: EntityTypeCodec
) -> langchain.tools.StructuredTool:
    def process_string(value: str) -> str:
        return anon_codec.anonymize(value, codec=codec)

    def wrap(function):
        if not function:
            return

        return wrap_with_output_processor(function, process_string=process_string)

    tool = tool.model_copy()
    tool.func = wrap(tool.func)
    tool.coroutine = wrap(tool.coroutine)
    return tool


def try_converting_to_openai_flex_tier(
    model,
    request_timeout: datetime.timedelta = datetime.timedelta(minutes=15),
):
    """The OpenAI Flex service tier is up to 10x cheaper than the Standard tier."""
    if not model.model_name.startswith("gpt-5") or model.model_name == "o4-mini":
        warnings.warn(f"Model {model.model_name} doesn't support Flex service tier.")
        return model

    model = model.model_copy(
        update={
            "service_tier": "flex",
            "request_timeout": request_timeout.total_seconds(),
        }
    )
    return model



import contextlib
import enum
import re
import string

import certifi
import gitlab as gl
import httpx
import pydantic

from ai_app.utils import PydanticForbidExtra, filter_dict_by_keys, try_match_regex, value_to_key


class GitlabUrl(enum.StrEnum):
    # Urls may have different permissions.
    gitlab = "https://gitlab.kapitalbank.az"
    dr_gitlab = "https://dr-gitlab.kapitalbank.az"


def build_gitlab_client(
    url: GitlabUrl = GitlabUrl.gitlab,
    private_token: str | None = None,
    verbose: bool = False,
) -> gl.Gitlab:
    gitlab = gl.Gitlab(
        url=url,
        private_token=private_token,
        ssl_verify=certifi.where(),
        keep_base_url=url != GitlabUrl.gitlab,
    )
    if verbose:
        gitlab.enable_debug()

    return gitlab


def build_gitlab_async_graphql_client(
    url: GitlabUrl = GitlabUrl.gitlab,
    private_token: str | None = None,
) -> gl.AsyncGraphQL:
    """
    Docs: https://docs.gitlab.com/api/graphql/
    Local instance GraphQL explorer: https://gitlab.kapitalbank.az/-/graphql-explorer
    """
    graphql = gl.AsyncGraphQL(
        url=url,
        token=private_token,
        ssl_verify=certifi.where(),
    )
    return graphql


def build_httpx_client(
    url: GitlabUrl = GitlabUrl.gitlab,
    private_token: str | None = None,
):
    headers = {"PRIVATE-TOKEN": private_token}
    client = httpx.AsyncClient(headers=headers, base_url=url)
    return client


class GraphqlGlobalId(PydanticForbidExtra):
    """
    GitLab GraphQL often uses global ids for object identifiers,
    see https://docs.gitlab.com/api/graphql/#global-ids
    """

    global_id: str
    schema_: str = pydantic.Field(alias="schema")
    provider: str
    path: list[str]
    entity_id: int

    @staticmethod
    def from_global_id(global_id: str):
        schema, path = global_id.split("://")
        provider, *path, entity_id = path.split("/")
        global_id = GraphqlGlobalId(
            global_id=global_id,
            schema=schema,
            provider=provider,
            path=path,
            entity_id=int(entity_id),
        )
        return global_id


async def get_project_id_from_path(graphql: gl.AsyncGraphQL, project: str) -> int:
    query = """
        query {
            project(fullPath: "$project") {
                id
            }
        }
    """
    result = await execute_parametrized_graphql_query(graphql, query, project=project)
    global_id = result["project"]["id"]
    global_id = GraphqlGlobalId.from_global_id(global_id)
    return global_id.entity_id


def get_gitlab_project_url_pattern() -> str:
    base_url_pattern = "|".join(re.escape(u) for u in GitlabUrl)
    pattern = rf"(?P<base_url>{base_url_pattern})/?(?P<project_path>[\w\-/]+)"
    return pattern


def get_gitlab_merge_request_url_pattern() -> str:
    pattern = rf"{get_gitlab_project_url_pattern()}/-/merge_requests/(?P<merge_request_iid>\d+)"
    return pattern


def gitlab_project_url_to_path(project_url: str) -> str:
    """
    For example, from
        https://gitlab.kapitalbank.az/kapitalbankojsc/birbank/birbank-payments/birbank-payments/ms-product-usage/
    to
        kapitalbankojsc/birbank/birbank-payments/birbank-payments/ms-product-usage
    """
    pattern = get_gitlab_project_url_pattern()
    match = try_match_regex(pattern, project_url, flags=re.ASCII)
    project_path = match.groupdict()["project_path"].strip("/")
    return project_path


class MergeRequest(pydantic.BaseModel):
    project_path: str
    merge_request_iid: int

    @staticmethod
    def from_url(merge_request_url: str):
        """For example, from https://gitlab.kapitalbank.az/kapitalbankojsc/birbank/birbank-payments/birbank-payments/ms-product-usage/-/merge_requests/4"""
        pattern = get_gitlab_merge_request_url_pattern()
        match = try_match_regex(pattern, merge_request_url, flags=re.ASCII)
        project_path = match.groupdict()["project_path"]
        merge_request_iid = int(match.groupdict()["merge_request_iid"])
        merge_request = MergeRequest(project_path=project_path, merge_request_iid=merge_request_iid)
        return merge_request


def get_merge_request_diffs(gitlab: gl.Gitlab, project: str, merge_request_iid: int) -> dict:
    project = gitlab.projects.get(project)
    merge_request = project.mergerequests.get(merge_request_iid)
    # https://docs.gitlab.com/api/merge_requests/#get-single-merge-request-changes
    changes = merge_request.changes(access_raw_diffs=True, unidiff=True)
    changes = [
        filter_dict_by_keys(change, ["old_path", "new_path", "diff"])
        for change in changes["changes"]
    ]
    merge_request = filter_dict_by_keys(merge_request.asdict(), ["title", "description"])
    merge_request |= {"changes": changes}
    return merge_request


async def aget_merge_request_diffs(
    gitlab: gl.Gitlab,
    project: str,
    merge_request_iid: int,
    graphql: gl.AsyncGraphQL | None = None,
    httpx_client: httpx.AsyncClient | None = None,
    access_raw_diffs: bool = True,
    unidiff: bool = True,
) -> dict:
    """https://docs.gitlab.com/api/merge_requests/#get-single-merge-request-changes"""
    gitlab_kwargs = dict(url=gitlab.url, private_token=gitlab.private_token)
    graphql = graphql or build_gitlab_async_graphql_client(**gitlab_kwargs)
    project_id = await get_project_id_from_path(graphql, project)

    changes_url = f"api/v4/projects/{project_id}/merge_requests/{merge_request_iid}/changes"
    if httpx_client:
        # Do not take responsibility to close httpx_client if it is provided.
        context = contextlib.nullcontext()
    else:
        httpx_client = context = build_httpx_client(**gitlab_kwargs)
    async with context:
        response = await httpx_client.get(
            changes_url, params=dict(access_raw_diffs=access_raw_diffs, unidiff=unidiff)
        )

    response.raise_for_status()
    merge_request = response.json()
    merge_request = filter_dict_by_keys(merge_request, ["title", "description", "changes"])
    merge_request["changes"] = [
        filter_dict_by_keys(change, ["old_path", "new_path", "diff"])
        for change in merge_request["changes"]
    ]
    return merge_request


async def execute_parametrized_graphql_query(
    graphql: gl.AsyncGraphQL, query: str, **parameters
) -> dict:
    query = string.Template(query).substitute(parameters)
    result = await graphql.execute(query)
    return result


async def get_all_schema_types(graphql: gl.AsyncGraphQL) -> dict[str, str]:
    query = """
        query {
            __schema {
                types {
                    name
                    kind
                }
            }
        }
    """
    result = await graphql.execute(query)
    types = value_to_key(result["__schema"]["types"], "name")
    return types


async def get_all_entity_type_fields(
    graphql: gl.AsyncGraphQL, entity_type: str, include_datatype_metadata: bool = False
) -> dict:
    """Note that entity_type is case-sensitive."""
    datatype_metadata = """
        type {
            name
            kind
            ofType {
                name
                kind
            }
        }
    """
    query = """
        query {
        __type(name: "$entity_type") {
            fields {
                name
                description
                $datatype_metadata
                args {
                    name
                    description
                    $datatype_metadata
                }
            }
        }
    }
    """
    result = await execute_parametrized_graphql_query(
        graphql,
        query,
        entity_type=entity_type,
        datatype_metadata=datatype_metadata if include_datatype_metadata else "",
    )
    fields = result["__type"]["fields"]
    for field in fields:
        field["args"] = value_to_key(field["args"], "name")

    fields = value_to_key(fields, "name")
    return fields


async def get_merge_request_commits(
    graphql: gl.AsyncGraphQL, project: str, merge_request_iid: int, first: int = 1
) -> list:
    query = """
        query {
            project(fullPath: "$project") {
                mergeRequest(iid: "$merge_request_iid") {
                    commits(first: $first) {
                        nodes {
                            title
                            message
                            diffs {
                                oldPath
                                newPath
                                diff
                            }
                        }
                    }
                }
            }
        }
    """
    result = await execute_parametrized_graphql_query(
        graphql, query, project=project, merge_request_iid=merge_request_iid, first=first
    )
    commits = result["project"]["mergeRequest"]["commits"]["nodes"]
    return commits



import textwrap
from typing import Annotated, override

import langchain.tools
import langsmith

from ai_app.utils import PydanticForbidExtra

from .base import BaseToolkit


@langchain.tools.tool()
def report_masked_entities_difficulties(
    masked_entities_difficulties: Annotated[
        str,
        textwrap.dedent("""
            Details regarding why the masked entities presented an issue and how the unmasked data 
            would have helped to improve the model response.
        """),
    ],
) -> str:
    """
    If the presence of masked entities like <IP_ADDRESS_0> in model input hindered the model's
    reasoning capabilities or degraded the quality of model response, then it should be reported
    with this function.
    """
    if run := langsmith.get_current_run_tree():
        run.add_tags("masked_entities_difficulties")
        run.add_metadata({"masked_entities_difficulties": masked_entities_difficulties})

    content = "Masked entities difficulties were reported."
    return content


@langchain.tools.tool()
def report_feedback(
    positive_feedback_and_recommendation: Annotated[
        str | None,
        textwrap.dedent("""
            Only include if the user expressed satisfaction, acknowledgment or gratitude.
            Extract and summarize specific aspects the user appreciated in the model response or user experience.
            Additionally list features or aspects of the interaction that worked well and should be maintained.
        """),
    ] = None,
    negative_feedback_and_recommendation: Annotated[
        str | None,
        textwrap.dedent("""
            Only include if the user expressed dissatisfaction, criticism or frustration.
            Identify and describe specific issues or shortcomings in the model response or user experience.
            Additionally describe how the user's experience can be improved,
            for example by modifying the model's system prompt, improving quality 
            of context or tools, modifying the web app interface or workflow, etc.
        """),
    ] = None,
) -> str:
    """
    Report user feedback or overall experience.
    Any messages with emotional tone, praise or criticism should be reported.
    """
    if run := langsmith.get_current_run_tree():
        metadata = {
            "positive_feedback_and_recommendation": positive_feedback_and_recommendation,
            "negative_feedback_and_recommendation": negative_feedback_and_recommendation,
        }
        metadata = {k: v for k, v in metadata.items() if v}
        if metadata:
            run.add_metadata(metadata)

    content = "Feedback and recommendation was reported."
    return content


class InfoSecUserReminder(PydanticForbidExtra):
    message: str


@langchain.tools.tool(response_format="content_and_artifact")
def report_disclosure_of_sensitive_information(
    information_type: Annotated[
        str,
        textwrap.dedent("""
            The type of disclosed sensitive information, such as password, 
            user or organisation client personal details, internal corporate data.
        """),
    ],
    artifact: Annotated[
        str,
        textwrap.dedent("""
            A reminder for the user that no sensitive information should be mentioned in this chat 
            and explaination why the information that the user provided is considered sensitive and 
            why it should not be shared.
        """),
    ],
) -> tuple[str, InfoSecUserReminder]:
    """
    If the user mentioned or unknowingly disclosed any sensitive or private information
    relating to the organisation customers or the organisation itself, then it must be
    reported using this function.
    """
    # Note that this logs to current run, not the root run (trace).
    if run := langsmith.get_current_run_tree():
        run.add_tags("user_disclosed_sensitive_information")
        run.add_metadata({"disclosed_sensitive_information_type": information_type})

    content = f"Disclosure of sensitive information type {information_type} was reported."
    artifact = InfoSecUserReminder(message=artifact)
    return content, artifact


class MetaToolkit(BaseToolkit):
    @override
    def get_tools(self) -> list[langchain.tools.BaseTool]:
        tools = [
            report_masked_entities_difficulties,
            report_disclosure_of_sensitive_information,
            report_feedback,
        ]
        return tools



import inspect
from typing import Iterable

import gradio as gr

from ai_app.external.providers import filter_common_model_names, get_common_models_meta_attributes


def build_model_choice_dropdown(
    models: Iterable[str] | None = None, max_cost: float = 5.0
) -> gr.Dropdown:
    common_models = get_common_models_meta_attributes()
    models = models or filter_common_model_names(max_cost=max_cost, supports_system_messages=True)
    choices = []
    for model in models:
        if attr := common_models.get(model):
            model = (f"{model} ({attr.get_relative_cost_representation()})", model)

        choices.append(model)

    model_choice = gr.Dropdown(choices, label="Model")
    return model_choice


def get_voice_activity_javascript():
    """
    Taken from https://www.gradio.app/guides/automatic-voice-detection#voice-activity-detection-for-hands-free-interaction
    This version detects only end of speech, and clicks the stop recording button,
    while the full version can detect start of speech also.
    """
    js = """
        async function main() {
            const script1 = document.createElement("script");
            script1.src = "https://cdn.jsdelivr.net/npm/onnxruntime-web@1.14.0/dist/ort.js";
            document.head.appendChild(script1);
            const script2 = document.createElement("script");
            script2.onload = async () =>  {
                console.log("vad loaded") ;
                var record = document.querySelector('.record-button');
                record.textContent = "Record"
                record.style = "width: fit-content; padding-right: 0.5vw;"
                const myvad = await vad.MicVAD.new(
                    {
                        onSpeechEnd: (audio) => {
                            var stop = document.querySelector('.stop-button');
                            if (stop != null) {
                                console.log(stop);
                                stop.click();
                            }
                        }
                    }
                );
                myvad.start();
            };
            script2.src = "https://cdn.jsdelivr.net/npm/@ricky0123/vad-web@0.0.7/dist/bundle.min.js";
            script1.onload = () =>  {
                console.log("onnx loaded") 
                document.head.appendChild(script2)
            };
        }
    """
    return js


def get_generating_text_javascript() -> str:
    js = inspect.cleandoc("""
    () => {
        // Function to start the animation on a specific element
        function startGeneratingAnimation(element) {
            // Ensure we don't attach multiple animations to the same element
            if (element.dataset.animated) return;
            element.dataset.animated = true;

            const animationFrames = ['Generating.', 'Generating..', 'Generating...'];
            let currentFrame = 0;
            
            const intervalId = setInterval(() => {
                // The most reliable way to stop is to check if the element has been
                // removed from the page. Gradio will replace it with the final response.
                if (!document.body.contains(element)) {
                    clearInterval(intervalId);
                    return;
                }

                element.textContent = animationFrames[currentFrame];
                currentFrame = (currentFrame + 1) % animationFrames.length;
            }, 400);
        }

        // Use a MutationObserver to watch for when our specific element is added to the DOM
        const observer = new MutationObserver((mutations) => {
            mutations.forEach((mutation) => {
                mutation.addedNodes.forEach((node) => {
                    // We only care about element nodes
                    if (node.nodeType !== 1) return;

                    // Check if the new node is our target element
                    if (node.id === 'gradio-generating-spinner') {
                        startGeneratingAnimation(node);
                    } 
                    // Also check if our target is a child of the new node
                    else {
                        const targetElement = node.querySelector('#gradio-generating-spinner');
                        if (targetElement) {
                            startGeneratingAnimation(targetElement);
                        }
                    }
                });
            });
        });

        // Start observing the entire document body for added nodes
        observer.observe(document.body, { childList: true, subtree: true });
    }
    """)
    return js


def get_generating_text_span() -> str:
    span = '<span id="gradio-generating-spinner" style="color: #888;">Generating...</span>'
    return span



import base64
import datetime
import enum
import functools
import importlib
import itertools
import logging
import os
import pathlib
import re
import site
import tempfile
import threading
import time
import urllib
import zlib
from types import ModuleType, TracebackType
from typing import Callable, Iterable, Literal, Self

import bidict
import cachetools.func
import fastapi
import gradio_client
import langchain_core.runnables
import limits.strategies
import numpy as np
import polars as pl
import pydantic
import urllib3
import uvicorn


def try_init_enum(enum_: type[enum.Enum], value) -> enum.Enum | None:
    try:
        return enum_(value)
    except ValueError:
        return


def filter_dict_by_keys(dictionary: dict, keys: Iterable) -> dict:
    dictionary = {key: dictionary[key] for key in keys if key in dictionary}
    return dictionary


def wrap_with_xml_tag(tag: str, value: str, with_new_lines: bool = False) -> str:
    value = f"\n{value}\n" if with_new_lines else value
    wrapped_value = f"<{tag}>{value}</{tag}>"
    wrapped_value += "\n" if with_new_lines else ""
    return wrapped_value


def normalize_whitespaces(string: str) -> str:
    string = re.sub(r"\s+", " ", string.strip())
    return string


def get_now() -> datetime.datetime:
    now = datetime.datetime.now().astimezone()
    return now


def get_utc_now() -> datetime.datetime:
    now = datetime.datetime.now(datetime.UTC)
    return now


class PydanticForbidExtra(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid")


def display_langchain_graph(
    graph_or_agent,
    endpoint: Literal["svg", "img"] = "svg",  # The img endpoint may hang.
    width: float = 500,
    **ipython_image_kwargs,
):
    """Displays graph as Mermaid diagram inside a Jupyter notebook."""
    import IPython.display

    graph = (
        graph_or_agent
        if isinstance(graph_or_agent, langchain_core.runnables.graph.Graph)
        else graph_or_agent.get_graph()
    )
    # Makes a call to Mermaid website, which may fail due to site rate limits.
    mermaid_syntax_string = graph.draw_mermaid()
    mermaid_syntax_string = base64.b64encode(mermaid_syntax_string.encode("utf8")).decode("ascii")
    url = f"https://mermaid.ink/{endpoint}/{mermaid_syntax_string}"
    image = IPython.display.Image(url=url, width=width, **ipython_image_kwargs)
    IPython.display.display(image)


def disable_urllib_insecure_request_warning():
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def get_number_of_batches(size: int, batch_size: int) -> int:
    n_batches = (size - 1) // batch_size + 1
    return n_batches


def unnest_with_prefix(df: pl.DataFrame, column: str) -> pl.DataFrame:
    df = df.with_columns(pl.col(column).struct.unnest().name.prefix(f"{column}_")).drop(column)
    return df


def refreshing_cache(
    maxsize: int | None = 128,
    ttl: float | datetime.timedelta = datetime.timedelta(seconds=60),
):
    """
    Calling a cached function with many different sets of arguments will lead to a memory leak.
    The cached function may return stale depending on function execution time or it the function
    throws an error after the first call.
    """
    ttl = ttl.total_seconds() if isinstance(ttl, datetime.timedelta) else ttl

    def decorator(function):
        @functools.lru_cache(maxsize=maxsize)
        class Cache:
            def __init__(self, *args, **kwargs):
                self.arguments = args, kwargs
                self.function_with_arguments = functools.partial(function, *args, **kwargs)
                self.value = self.function_with_arguments()
                self.thread = None
                self.lock = threading.Lock()

            def update(self):
                try:
                    self.value = self.function_with_arguments()
                except Exception as e:
                    logging.error(
                        f"Encountered an error when trying to refresh function cache for '{function}' with arguments {self.arguments}:\n{e}",
                        exc_info=True,
                    )

            def is_thread_running(self):
                return self.thread and self.thread.is_alive()

            def start_thread_if_not_running(self):
                with self.lock:
                    if not self.is_thread_running():
                        self.thread = threading.Thread(target=self.update, daemon=True)
                        self.thread.start()

            @cachetools.func.ttl_cache(ttl=ttl)
            def get_value(self):
                self.start_thread_if_not_running()
                return self.value

        @functools.wraps(function)
        def wrapped(*args, **kwargs):
            cache = Cache(*args, **kwargs)
            value = cache.get_value()
            return value

        return wrapped

    return decorator


class LazyImporter:
    """
    Useful for heavy packages. Example:

        os = LazyImporter("os")
        os.getcwd()
    """

    def __init__(self, module_name: str):
        # Long name not to collide with module namespace because of getattr method.
        self._module_name_for_lazy_import = module_name

    @functools.cache
    def _get_cached_lazy_import_module(self):
        with Timer("lazy importing"):
            module = importlib.import_module(self._module_name_for_lazy_import)

        return module

    def __getattr__(self, attribute: str):
        module = self._get_cached_lazy_import_module()
        attribute = getattr(module, attribute)
        return attribute


class DecoratorInterface:
    callable: Callable = None

    def on_enter(self) -> None:
        """On enter hook."""

    def on_exit(self) -> None:
        """On exit hook."""

    def on_before_callable(self, *args, **kwargs) -> None:
        """A hook, args and kwargs are the ones that were passed to the function."""

    def on_after_callable(self, result, *args, **kwargs) -> None:
        """A hook, function returned result when it was called with args and kwargs."""

    def decorate(self, callable: Callable) -> Callable:
        @functools.wraps(callable)
        def wrap(*args, **kwargs):
            self.callable = callable
            self.on_enter()
            self.on_before_callable(*args, **kwargs)
            result = callable(*args, **kwargs)
            self.on_after_callable(result, *args, **kwargs)
            self.on_exit()
            self.callable = None
            return result

        return wrap

    __call__ = decorate


class Timer(DecoratorInterface):
    """Context manager and decorator for timing."""

    def __init__(self, name: str | None = None, warning_threshold: float | None = None):
        self.name = name
        self.warning_threshold = float("inf") if warning_threshold is None else warning_threshold
        self.begin: float | None = None
        self.end: float | None = None

    def get_enriched_name(self):
        name_parts = []
        if self.name:
            name_parts.append(self.name)
        if self.callable is not None:
            name_parts.append(f"callable {self.callable.__qualname__}")

        name = ", ".join(name_parts)
        return name

    @property
    def delta(self) -> float | None:
        return None if self.end is None else self.end - self.begin

    def on_enter(self):
        self.begin = time.time()
        self.end = None
        logging.info(self.get_enriched_name())

    def on_exit(self):
        self.end = time.time()
        level = logging.WARNING if self.delta > self.warning_threshold else logging.INFO
        logging.log(level, f"Time spent on {self.get_enriched_name()}:\t{self.delta:.6f} s")

    def __enter__(self):
        self.on_enter()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_val: BaseException | None = None,
        exc_tb: TracebackType | None = None,
    ):
        self.on_exit()


def setup_logging(level=logging.INFO, **kwargs):
    logging.basicConfig(
        level=level, format="{asctime} | {levelname} | {message}", style="{", **kwargs
    )


def try_match_regex(pattern: str, string: str, **kwargs):
    match = re.match(pattern, string, **kwargs)
    if not match:
        raise ValueError(f"String '{string}' does not fit expected pattern '{pattern}'")

    return match


def int_log_2(value: float) -> int:
    value = int(np.log2(value))
    return value


def get_pydantic_model_fields(model: pydantic.BaseModel) -> dict[str]:
    """Preserves field types instead of converting to base Python types, like model_dump does."""
    model_fields = {field: getattr(model, field) for field in sorted(model.model_fields_set)}
    return model_fields


def get_total_disk_size(path: str = ".") -> float:
    """Calculates the total size of a file or directory"""
    if os.path.isfile(path):
        size = os.path.getsize(path)
        return size

    if not os.path.isdir(path):
        raise ValueError(f"Path '{path}' is not a valid file or directory.")

    size = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            if os.path.islink(filepath):
                continue

            try:
                size += os.path.getsize(filepath)
            except OSError as e:
                print(f"Could not get size of {filepath}: {e}")

    return size


def get_installed_packages_sizes() -> pl.DataFrame:
    packages = []
    for site_package in site.getsitepackages():
        names = os.listdir(site_package)
        for name in names:
            path = os.path.join(site_package, name)
            if os.path.isdir(path) and not path.endswith(".dist-info"):
                package_size = get_total_disk_size(path)
                packages = {"package": name, "size_mb": package_size / 10**6, "package_dir": path}
                packages.append(packages)

    packages = pl.DataFrame(packages).sort("size_mb", descending=True)
    return packages


class CallableCounter:
    """Returns an increasing integer every time it is called."""

    def __init__(self):
        self.count = itertools.count()

    def __call__(self) -> int:
        return next(self.count)


class CounterBidict(bidict.bidict):
    """A bidict which maps to an autoincreasing index."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._callable_counter = CallableCounter()

    def __getitem__(self, key):
        if key not in self:
            self[key] = self._callable_counter()

        value = super().__getitem__(key)
        return value


def value_to_key(items: list, key: str) -> dict:
    dictionary = {i.pop(key): i for i in items}
    return dictionary


def join_url(base_url: str, *parts: str, add_trailing_slash: bool = False) -> str:
    url = base_url.rstrip("/")
    for part in parts:
        url = urllib.parse.urljoin(url + "/", part.strip("/"))

    url = url + "/" if add_trailing_slash else url
    return url


async def serve_fastapi_app(app: fastapi.FastAPI, port: int = 7860, **kwargs):
    """Intended to be used in Jupyter notebooks."""
    config = uvicorn.Config(app, port=port, **kwargs)
    server = uvicorn.Server(config)
    await server.serve()


def build_url_with_parameters(base_url: str, **params) -> str:
    if not params:
        return base_url

    url = f"{base_url}?{urllib.parse.urlencode(params)}"
    return url


def file_to_base64(path: str) -> str:
    with open(path, "rb") as f:
        base64_file = base64.b64encode(f.read()).decode("utf-8")

    return base64_file


def compress(data: str) -> str:
    data_bytes = zlib.compress(data.encode("utf-8"))
    compressed_data = base64.b64encode(data_bytes).decode("utf-8")
    return compressed_data


def decompress(compressed_data: str) -> str:
    data_bytes = base64.b64decode(compressed_data)
    data = zlib.decompress(data_bytes).decode("utf-8")
    return data


def get_module_path(module: ModuleType) -> pathlib.Path:
    path = importlib.resources.files(module)
    return path


def fetch_gradio_named_endpoints_parameter_labels(
    client: gradio_client.Client,
) -> dict[str, dict[str, str]]:
    api_info = client.view_api(print_info=False, return_format="dict")
    endpoints = api_info["named_endpoints"]
    endpoints = {
        k: {p["label"]: p["parameter_name"] for p in v["parameters"]} for k, v in endpoints.items()
    }
    return endpoints


def is_redirect_response(response: fastapi.Response) -> bool:
    return 300 <= response.status_code < 400 and "Location" in response.headers


def generate_test_file(size_mb: int) -> pathlib.Path:
    """Generate a test file of the specified size in MB."""
    temp_dir = tempfile.gettempdir()
    file_path = pathlib.Path(temp_dir) / f"test_{size_mb}mb.bin"
    mega = 1024 * 1024
    chunk = b"0" * mega
    with open(file_path, "wb") as f:
        for _ in range(size_mb):
            f.write(chunk)

    actual_size_mb = os.path.getsize(file_path) / mega
    assert abs(actual_size_mb - size_mb) < 0.1
    return file_path


class HttpMethod(enum.StrEnum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    HEAD = "HEAD"
    OPTIONS = "OPTIONS"


def get_remote_address(request: fastapi.Request) -> str:
    address = request.client.host if request.client and request.client.host else "127.0.0.1"
    return address


def are_rate_limits_hit(
    limiter: limits.strategies.RateLimiter,
    limits_with_keys: Iterable[tuple[limits.RateLimitItem, tuple[str, ...]]],
) -> bool:
    """Need to eagerly iterate over all rate limits and keys combinations to ensure that all are hit."""
    hit = False
    for rate_limit, key in limits_with_keys:
        allow = limiter.hit(rate_limit, *key)
        hit = hit or not allow

    return hit



import inspect
from typing import override

import gradio as gr
import langchain_core.language_models
import langchain_core.messages

from ai_app.apps.merge_request_review.backend import Bot
from ai_app.core import BaseApp, Response, State


def get_examples():
    examples = [
        """
            https://gitlab.kapitalbank.az/kapitalbankojsc/birbank/birbank-bff/bff-payment-product/-/merge_requests/150
        """,
        """
            https://gitlab.kapitalbank.az/kapitalbankojsc/birbank/birbank-payments/birbank-payments/ms-product-usage/-/merge_requests/4
            https://gitlab.kapitalbank.az/kapitalbankojsc/birbank/consul-config-loader/birbank-payments/consul-config-loader/-/merge_requests/420
        """,
    ]
    examples = [[inspect.cleandoc(e)] for e in examples]  # Convert to Gradio examples format.
    return examples


class App(BaseApp):
    name = "Merge request review"
    requires_auth = False

    def __init__(self):
        self.bot = Bot()

    @override
    async def astream(
        self,
        model: langchain_core.language_models.BaseChatModel,
        message: langchain_core.messages.BaseMessage,
        frontend_messages: list[langchain_core.messages.BaseMessage],
        state: State,
    ):
        messages = state.messages + [message]
        async for response in self.bot.astream(model, messages, state):
            yield Response(content=response.formatted_response, messages=response.state["messages"])

    @override
    def build_gradio_blocks(self, model_choice: gr.Dropdown | None = None) -> gr.Blocks:
        with gr.Blocks(fill_height=True) as app:
            self.build_gradio_chat_interface(
                model_choice=model_choice,
                description="To start the review, provide a GitLab merge request url.",
                examples=get_examples(),
            )

        return app



import inspect

import altair
import langchain.tools
import langchain_core.language_models
import langchain_core.messages
import langgraph.prebuilt
import pandas as pd

from ai_app.ai_utils import PreModelHook
from ai_app.config import get_config, get_engine, get_sql_toolkit
from ai_app.core import Response, State
from ai_app.tools import extract_tool_artifacts_from_agent_response


def get_system_prompt() -> str:
    prompt = inspect.cleandoc("""
    You are an agent designed to interact with a SQL database.
    Given an input question, create a syntactically correct query to run, then look 
    at the results of the query and return the answer.
    You can order the results by a relevant column to return the most interesting examples in the database.
    Never query for all the columns from a specific table, only ask for the relevant columns given the question.
    You have access to tools for interacting with the database.
    Only use the below tools. Only use the information returned by the below tools to construct your final answer.
    You MUST double check your query before executing it. If you get an error while executing a query, rewrite 
    the query and try again.

    DO NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the database.

    To start you should ALWAYS look at the tables in the database to see what you can query.
    Do NOT skip this step.
    Then you should query the schema of the most relevant tables.

    When referring to table columns, always quote them with double quotes, like "Metric" or "Date".
    """)  # noqa: E501
    return prompt


class Bot:
    async def respond(
        self,
        model: langchain_core.language_models.BaseChatModel,
        message: langchain_core.messages.BaseMessage,
        frontend_messages: list[langchain_core.messages.BaseMessage],
        state: State,
        connection_context_name: str,
    ) -> Response:
        connection_context = get_config().sql_connection_contexts[connection_context_name]
        engine = get_engine(connection_context.connection_name)

        @langchain.tools.tool(response_format="content_and_artifact")
        def display_altair_chart(
            vega_lite_json_spec: str,
            sql_query: str,
        ) -> tuple[str, str]:
            """
            Display a single chart from a Vega-Lite JSON specification and a SQL query.
            The connection name can be determined by calling the get_database_info tool of an SQL MCP.
            Refer to data source as "sql_query", for example:
            ```json
            {
            "$schema": "https://vega.github.io/schema/vega-lite/v6.json",
            "data": {"name": "sql_query"},
            "title": "My Bar Chart",
            "mark": "bar",
            "encoding": {
                "x": {"field": "category", "type": "nominal", "title": "Category"},
                "y": {"field": "value", "type": "quantitative", "title": "Value"},
                "color": {"field": "group", "type": "nominal", "title": "Group"}
            },
            "width": 400,
            "height": 300
            }
            ```
            Spaces in the example JSON are just for human readability, when generating the real JSON,
            omit all spaces.

            Column names in both the JSON spec and the SQL query should be lowercase and quoted.
            """
            # Reference: https://vega.github.io/vega-lite
            sql_query = sql_query.strip().strip(";")
            df = pd.read_sql(sql_query, engine)
            chart = altair.Chart.from_json(vega_lite_json_spec)
            chart.data = df

            response = "HTML displayed."
            return response, chart

        toolkit = get_sql_toolkit(connection_context_name)
        agent = langgraph.prebuilt.create_react_agent(
            model=model,
            pre_model_hook=PreModelHook(model),
            tools=toolkit.get_tools() + [display_altair_chart],
            prompt=get_system_prompt(),
        )
        response: dict = await agent.ainvoke({"messages": frontend_messages + [message]})
        response_messages = response["messages"]
        response_content = response_messages[-1].content

        artifacts = extract_tool_artifacts_from_agent_response(response_messages)
        artifacts = artifacts.get(display_altair_chart.name)
        plot = artifacts[-1] if artifacts else None

        return Response(content=response_content, additional_outputs=[plot])



import logging

import fastapi.middleware.cors
import fastapi.responses
import starlette.middleware.base
import starlette.middleware.sessions
import starlette.types

from ai_app.config import RateLimitConfig, get_config
from ai_app.utils import is_redirect_response


def add_middleware(app: fastapi.FastAPI):
    """
    Order of middleware matters - order of dispatch calls is reverse to the order of middleware addition.
    So the last added middleware is first to see the request and last to see the response.
    """
    app.add_middleware(
        RateLimitsMiddleware,
        rate_limit_config=get_config().rate_limit_config,
    )
    app.add_middleware(
        starlette.middleware.sessions.SessionMiddleware,
        secret_key=get_config().secrets.session_crypto_key.get_secret_value(),
    )
    app.add_middleware(BlockGradioFileEndpointRedirectsMiddleware)
    app.add_middleware(
        SecurityHeadersMiddleware,
        security_headers=get_config().security_headers,
    )
    app.add_middleware(RemoveGradioCorsHeadersMiddleware)
    app.add_middleware(
        fastapi.middleware.cors.CORSMiddleware,
        **get_config().cors_config.model_dump(),
    )
    if get_config().is_prod:
        app.add_middleware(SanitizeInternalErrorResponsesMiddleware)


def try_log_response_body_as_error(response: fastapi.Response):
    try:
        body = response.body
        body = body.tobytes() if isinstance(body, memoryview) else body
        body = body.decode()
        logging.error(f"Internal server error: {body}")
    except Exception as e:
        logging.error(f"Error processing error response: {e}")


def get_generic_500_error_response() -> fastapi.responses.JSONResponse:
    generic_500_error_response = fastapi.responses.JSONResponse(
        status_code=500,
        content={"detail": "Internal server error"},
    )
    return generic_500_error_response


def get_generic_429_error_response() -> fastapi.responses.JSONResponse:
    generic_429_error_response = fastapi.responses.JSONResponse(
        status_code=429,
        content={"detail": "Rate limit exceeded"},
    )
    return generic_429_error_response


class SanitizeInternalErrorResponsesMiddleware(starlette.middleware.base.BaseHTTPMiddleware):
    async def dispatch(self, request: fastapi.Request, call_next) -> fastapi.Response:
        try:
            response = await call_next(request)
        except Exception as e:
            logging.error(f"Error when processing request: {e}")
            response = get_generic_500_error_response()
        else:
            if response.status_code >= 500:
                response = get_generic_500_error_response()
                try_log_response_body_as_error(response)

        return response


class BlockGradioFileEndpointRedirectsMiddleware(starlette.middleware.base.BaseHTTPMiddleware):
    async def dispatch(self, request: fastapi.Request, call_next) -> fastapi.Response:
        response = await call_next(request)
        if is_redirect_response(response) and "/gradio_api/file=" in request.url.path:
            response = fastapi.Response(
                status_code=400,
                content="Redirects are not allowed for Gradio file endpoint",
            )

        return response


class RemoveGradioCorsHeadersMiddleware(starlette.middleware.base.BaseHTTPMiddleware):
    """
    Removes Gradio CORS headers from the response, as they may be too permissive.
    For enhanced safety, the headers in the response should be properly set after this middleware.
    """

    async def dispatch(self, request: fastapi.Request, call_next) -> fastapi.Response:
        response = await call_next(request)
        del response.headers["Access-Control-Allow-Origin"]
        del response.headers["Access-Control-Allow-Credentials"]
        return response


class SecurityHeadersMiddleware(starlette.middleware.base.BaseHTTPMiddleware):
    def __init__(self, app: starlette.types.ASGIApp, security_headers: dict[str, str]):
        """The app argument is the FastAPI app or the previous middleware instance in the chain."""
        super().__init__(app)
        self.security_headers = security_headers

    async def dispatch(self, request: fastapi.Request, call_next) -> fastapi.Response:
        response = await call_next(request)
        response.headers.update(self.security_headers)
        return response


class RateLimitsMiddleware(starlette.middleware.base.BaseHTTPMiddleware):
    """
    A SessionMiddleware should be added after this middleware
    if the rate limit key depends on the session.
    """

    def __init__(self, app: starlette.types.ASGIApp, rate_limit_config: RateLimitConfig):
        super().__init__(app)
        self.limiter = rate_limit_config.build_limiter()
        self.rate_limit_config = rate_limit_config

    async def dispatch(self, request: fastapi.Request, call_next):
        if self.rate_limit_config.are_rate_limits_hit(request, self.limiter):
            response = get_generic_429_error_response()
        else:
            response = await call_next(request)

        return response
